#ifndef ERMYTH_FACTORY_H
#define ERMYTH_FACTORY_H

#include <string>
#include <map>

#include <functional>

// i don't like this dep, but i also don't like
// what happens in some cases when i don't use
// phoenix. :/
#include <ermyth/phoenix.h>

namespace impl
{
  template <typename T>
  inline T *null_factory ()
  {
    return 0;
  }
}

namespace factory
{
  template<class T, class SubT>
  struct create_hook
  {
    typedef T *         result_type;
    typedef SubT        actual_type;

    static result_type create ()
    {
      return new actual_type;
    }

    result_type operator () () const
    {
      return create ();
    }
  };


  template<typename AliasedType>
  struct aliaser
  {
    typedef AliasedType key_type;
    typedef AliasedType value_type;

    // A map type for storing lookup key aliases.
    typedef std::map<key_type, value_type> alias_map_type;

    // Aliases 'alias' as an equivalent of 'isthesameas'.
    void alias (const key_type &analias, const value_type &isthesameas)
    {
      aliases () [analias] = isthesameas;
    }

    // Returns the map of aliases.
    alias_map_type &aliases ()
    {
      return m_map;
    }

    const alias_map_type &aliases () const
    {
      return m_map;
    }

    typedef typename alias_map_type::iterator iterator;
    typedef typename alias_map_type::const_iterator const_iterator;


    /**
     * Iterators for aliases ().
     */

    iterator begin ()
    {
      return m_map.begin ();
    }

    const_iterator begin () const
    {
      return m_map.begin ();
    }

    iterator end ()
    {
      return m_map.end ();
    }

    const_iterator end () const
    {
      return m_map.end ();
    }

    /**
     * Expands the given alias recursively. If a circular
     * alias is detected, the last expansion is returned
     * (i.e., same as _alias). If no expansion is found
     * then _alias is returned.
     */
    value_type expand (const key_type &_alias) const
    {
      typename alias_map_type::const_iterator cit = m_map.find (_alias);
      typename alias_map_type::const_iterator cet = m_map.end ();

      if (cet == cit)
        return _alias;

      value_type exp = (*cit).second;

      while (true)
        {
          cit = m_map.find (exp);
          if (cet == cit)
            return exp;
          exp = (*cit).second;
          if (exp == _alias)
            return exp;
        }

      return exp;
    }

  private:
    alias_map_type m_map;
  };
  
  /**
   * Example:
   *
   * factory_mgr<MyInterface> &fac = factory_mgr<MyInterface>::instance ();
   * fac.register_factory ("my_key", factory::create_hook<MyInterface,MyClass>::create);
   * MyInterface *foo = fac.create ("some_key"); // == NULL
   * foo = fac.create ("my_key"); // == a new MyClass object
   */
  template<class InterfaceT, class KeyType = std::string>
  class factory_mgr
  {
  public:
    // typedefs for conformance with the Adaptable Unary Functor model
    typedef KeyType key_type;
    typedef InterfaceT value_type;
    typedef InterfaceT *result_type;

    /**
     * The type of factories used by this class: a
     * function taking void and returning (value_type *).
     * See factory_map ().
     * 
     * todo: implement proper functor support.
     */
    typedef result_type (*factory_function_type) ();

    /**
     * Internal container type used for mapping keys to
     * factories.
     */
    typedef std::map<key_type, factory_function_type> factory_map_type;

    /**
     * A type used to map classname aliases for
     * this factory manager.
     */
    typedef aliaser<key_type> aliaser_type;

  private:
    // Internal marker type.
    struct context
    {
    };

  protected:
    factory_map_type &factory_map ()
    {
      return ::impl::phoenix<factory_map_type, factory_mgr>::instance ();
    }

    const factory_map_type &factory_map () const
    {
      return ::impl::phoenix<factory_map_type, factory_mgr>::instance ();
    }

    aliaser_type &aliases ()
    {
      return ::impl::phoenix<aliaser_type, factory_mgr>::instance ();
    }
    
  public:
    factory_mgr ()
    {
    }

    virtual ~factory_mgr () throw ()
    {
    }

    virtual result_type create (const key_type &_key)
    {
      factory_function_type func = 0;
      {
        key_type key = aliases ().expand (_key);
        typename factory_map_type::const_iterator it = factory_map ().find (key);
        if (it != factory_map ().end ())        // found a factory?
          {
            func = it->second;
          }
      }
      return (0 == func)
        ? 0
        : func ();
    }

    result_type operator () (const key_type &key)
    {
      return create (key);
    }

    virtual void destroy (result_type obj)
    {
      delete obj;
    }

    virtual void register_factory (const key_type &key, factory_function_type fp)
    {
      factory_map ().insert (typename factory_map_type::value_type (key, fp));
    }

    virtual bool provides (const key_type &key) const
    {
      return factory_map ().end () != factory_map ().find (aliases ().expand (key));
    }

    static factory_mgr &instance ()
    {
      return ::impl::phoenix<factory_mgr>::instance ();
    }

    const aliaser_type &aliases () const
    {
      return ::impl::phoenix<aliaser_type, factory_mgr>::instance ();
    }

    void alias (std::string const &Alias, std::string const &Real)
    {
      aliases ().alias (Alias, Real);
    }

  }; // class factory_mgr

  template <typename InterfaceT>
  inline factory_mgr<InterfaceT, std::string> &
  factory ()
  {
    return factory_mgr<InterfaceT, std::string>::instance ();
  }

  template <typename InterfaceT>
  inline void
  register_factory (const std::string &classname, InterfaceT *(*factory_function) ())
  {
    factory<InterfaceT> ().register_factory (classname, factory_function);
  }

  template <typename InterfaceT>
  inline void
  register_factory (const std::string &classname)
  {
    factory<InterfaceT> ().register_factory (classname, create_hook<InterfaceT, InterfaceT>::create);
  }

  template <typename InterfaceT, typename ImplT>
  inline void
  register_subtype (const std::string &classname)
  {
    factory<InterfaceT> ().register_factory (classname, create_hook<InterfaceT, ImplT>::create);
  }

  template <typename InterfaceT>
  void register_abstract (const std::string &classname)
  {
    factory<InterfaceT> ().register_factory (classname, ::impl::null_factory<InterfaceT>);
  }

  template <typename InterfaceT>
  inline InterfaceT *
  create (const std::string &classname)
  {
    return factory<InterfaceT> ().create (classname);
  }

  template <typename InterfaceT>
  inline void
  alias (std::string const &Alias, const std::string &Real)
  {
    return factory<InterfaceT> ().alias (Alias, Real);
  }
} // namespace factory

#endif // ERMYTH_FACTORY_H
