/**
 * base64.h: Base64 encoding / decoding algorithm.
 *
 * Copyright © 2007 Pippijn van Steenhoven / The Ermyth Team
 * Rights to this code are as documented in COPYING.
 *
 * This is a modified version of libb64 (http://sourceforge.net/projects/libb64)
 *
 * $Id: base64.h,v 1.1 2007/09/22 14:27:26 pippijn Exp $
 */

#ifndef BASE64_H
#define BASE64_H

#include <iostream>

namespace base64
{
  enum base64_decodestep
  {
    step_a, step_b, step_c, step_d
  };

  struct base64_decodestate
  {
    base64_decodestep step;
    char plainchar;
  };

  void base64_init_decodestate (base64_decodestate *state_in);

  int base64_decode_value (char value_in);

  int base64_decode_block (const char *code_in, const int length_in, char *plaintext_out, base64_decodestate *state_in);

  struct decoder
  {
    base64_decodestate _state;
    int _buffersize;

    decoder (int buffersize_in = 4096):_buffersize (buffersize_in)
    {
    }
    int decode (char value_in)
    {
      return base64_decode_value (value_in);
    }
    template<int length_in>
    int decode (const char (&code_in)[length_in], char *plaintext_out)
    {
      return decode (code_in, length_in, plaintext_out);
    }
    int decode (const char *code_in, const int length_in, char *plaintext_out)
    {
      base64_init_decodestate (&_state);
      return base64_decode_block (code_in, length_in, plaintext_out, &_state);
    }
    int decode (char const *code, const int ilen, char **otxt, size_t *tlen)
    {
      int olen = 1 + (((ilen + 2) / 3) * 4);
      if (ilen > olen)
        {
          *otxt = NULL;
          return 0;
        }

      *otxt = new char[olen];

      if (*otxt)
        decode (code, ilen, *otxt);
      else
        throw std::bad_alloc ();

      *tlen = olen;

      return olen - 1;
    }
    void decode (std::istream & istream_in, std::ostream & ostream_in)
    {
      base64_init_decodestate (&_state);
      //
      const int N = _buffersize;
      char *code = new char[N];
      char *plaintext = new char[N];
      int codelength;
      int plainlength;

      do
	{
	  istream_in.read ((char *) code, N);
	  codelength = istream_in.gcount ();
	  plainlength = decode (code, codelength, plaintext);
	  ostream_in.write ((const char *) plaintext, plainlength);
	}
      while (istream_in.good () && codelength > 0);
      //
      base64_init_decodestate (&_state);

      delete[]code;
      delete[]plaintext;
    }
  };

}				// namespace base64

namespace base64
{
  enum base64_encodestep
  {
    step_A, step_B, step_C
  };

  struct base64_encodestate
  {
    base64_encodestep step;
    char result;
    int stepcount;
  };

  void base64_init_encodestate (base64_encodestate *state_in);

  char base64_encode_value (char value_in);

  int base64_encode_block (const char *plaintext_in, int length_in, char *code_out, base64_encodestate *state_in);

  int base64_encode_blockend (char *code_out, base64_encodestate *state_in);

  struct encoder
  {
    base64_encodestate _state;
    int _buffersize;

    encoder (int buffersize_in = 4096):_buffersize (buffersize_in)
    {
    }
    int encode (char value_in)
    {
      return base64_encode_value (value_in);
    }
    template<int length_in>
    int encode (const char (&code_in)[length_in], char *plaintext_out)
    {
      return encode (code_in, length_in, plaintext_out);
    }
    int encode (const char *code_in, const int length_in, char *plaintext_out)
    {
      base64_init_encodestate (&_state);
      return base64_encode_block (code_in, length_in, plaintext_out, &_state);
    }
    int encode_end (char *plaintext_out)
    {
      return base64_encode_blockend (plaintext_out, &_state);
    }
    template<int length_in>
    int encode (char const (&code)[length_in], char **otxt)
    {
      return encode (code, length_in, otxt);
    }
    int encode (char const * const code, const int ilen, char **otxt)
    {
      int olen = 1 + (((ilen + 2) / 3) * 4);
      if (ilen > olen)
        {
          *otxt = NULL;
          return 0;
        }

      *otxt = new char[olen];

      if (*otxt)
        encode (code, ilen, *otxt);
      else
        throw std::bad_alloc ();

      return olen - 1;
    }
    void encode (std::istream &istream_in, std::ostream &ostream_in)
    {
      base64_init_encodestate (&_state);
      //
      const int N = _buffersize;
      char *plaintext = new char[N];
      char *code = new char[2 * N];
      int plainlength;
      int codelength;

      do
	{
	  istream_in.read (plaintext, N);
	  plainlength = istream_in.gcount ();
	  //
	  codelength = encode (plaintext, plainlength, code);
	  ostream_in.write (code, codelength);
	}
      while (istream_in.good () && plainlength > 0);

      codelength = encode_end (code);
      ostream_in.write (code, codelength);
      //
      base64_init_encodestate (&_state);

      delete[]code;
      delete[]plaintext;
    }
  };

}				// namespace base64

#endif // BASE64_H
