/**
 * util.h: Various utility functions.
 *
 * Copyright © 2007 Pippijn van Steenhoven / The Ermyth Team
 * Rights to this code are as documented in COPYING.
 *
 * $Id: containers.h,v 1.1 2007/09/22 14:27:26 pippijn Exp $
 */

#ifndef UTIL_H
#define UTIL_H

#include <climits>
#include <cstddef>
#include <vector>

/**
 * Mostly the same as std::vector, but erase can reorder
 * the elements, making insert/remove O(1) instead of O(n).
 *
 * NOTE: only some forms of erase are available
 * (Taken from Crossfire TRT http://cf.schmorp.de/)
 */
template<typename T, typename Alloc = std::allocator<T> >
struct unordered_vector : std::vector<T, Alloc>
{
  typedef typename unordered_vector::iterator iterator;

  void erase (unsigned pos)
  {
    if (pos < this->size () - 1)
      (*this)[pos] = (*this)[this->size () - 1];

    this->pop_back ();
  }

  void erase (iterator pos)
  {
    erase ((unsigned)(pos - this->begin ()));
  }

  void insert (T const &val)
  {
    push_back (val);
  }
};

// This is a vector that stores each element's index in the object.
// This only works for pointers to objects.
template<class T, unsigned T::*index = &T::index, class Alloc = std::allocator<T *> >
struct indexing_vector : std::vector<T *, Alloc>
{
  typedef typename indexing_vector::size_type size_type;
  typedef typename indexing_vector::iterator iterator;
  
  iterator find (const T *obj)
  {
    return obj->*index
      ? this->begin () + obj->*index - 1
      : this->end ();
  }

  void erase (T *obj)
  {
    unsigned pos = obj->*index;
    obj->*index = 0;

    if (pos < this->size ())
      {
        (*this)[pos - 1] = (*this)[this->size () - 1];
        (*this)[pos - 1]->*index = pos;
      }

    this->pop_back ();
  }

  void insert (T *obj)
  {
    this->push_back (obj);
    obj->*index = this->size ();
  }
};

template<typename T, typename VectorT>
inline typename VectorT::iterator
find_element (T *element, VectorT &vec)
{
  typename VectorT::iterator it = vec.begin ();
  typename VectorT::iterator et = vec.end ();

  while (it != et)
    {
      if (*it == element)
        return it;
      ++it;
    }

  return et;
}

/**
 * The following templates find a sufficiently large integer type to contain
 * the given number of bits.
 *
 * Usage:
 *
 * unsigned_type<16>::type a; // 16-bit unsigned if available (probably unsigned short)
 * signed_type<8>::type a; // 8-bit signed if available (probably char)
 */

// Empty class if everything failed
struct empty_base
{
};

// To allow inheritance from this and get the integer type out
template<typename T>
struct identity
{
  typedef T type;
};

/**
 * The following two templates are a compile-time if-statement.
 */
template<bool Cond, class Then, class Else>
struct If
{
  typedef Then ret;
};

template<class Then, class Else>
struct If<false, Then, Else>
{
  typedef Else ret;
};

// Looks whether the desired type exists
template<typename T, std::size_t bits, typename Else>
struct type_if_size
: If<sizeof (T) * CHAR_BIT >= bits, identity<T>, Else>::ret
{
};

template<std::size_t bits>
struct unsigned_type
: type_if_size<unsigned char, bits,
    type_if_size<unsigned short, bits,
      type_if_size<unsigned int, bits,
        type_if_size<unsigned long, bits,
          empty_base // No such type
        >
      >
    >
  >
{
};

template<std::size_t bits>
struct signed_type
: type_if_size<signed char, bits
  , type_if_size<signed short, bits
    , type_if_size<signed int, bits
      , type_if_size<signed long, bits
        , empty_base // No such type
        >
      >
    >
  >
{
};

#endif
