/**
 * exception.h: Ermyth exceptions.
 *
 * Copyright © 2007 Pippijn van Steenhoven / The Ermyth Team
 * Rights to this code are as documented in doc/pod/gplicense.pod.
 *
 * $Id: exception.h,v 1.1 2007/09/22 14:27:26 pippijn Exp $
 */

#ifndef ERMYTH_EXCEPTION_H
#define ERMYTH_EXCEPTION_H

#include <cstdarg>
#include <cstdio>
#include <algorithm>
#include <exception>

#include <svsconfig.h>

#define SRCINF  ::impl::srcinf (__FILE__, __LINE__, CURFUNC)

#define EBUFSIZE 1024

namespace impl
{
   struct srcinf
   {
     srcinf (char const *filename, unsigned int linenumber, char const *function)
     : m_info (info (filename, linenumber, function))
     {
     }
   
     unsigned int line() const throw() { return m_info.line; }
     char const *file() const throw()  { return m_info.file; }
     char const *func() const throw()  { return m_info.func; }
     
     srcinf &operator = (srcinf const &rhs)
     {
       if (this != &rhs)
         this->m_info = rhs.m_info;
       return *this;
     }
   
     srcinf (srcinf const &rhs)
     {
       this->m_info = rhs.m_info;
     }
   
   private:
     struct info
     {
       char const *file;
       unsigned line;
       char const *func;
   
       info ()
       : file (0), line (0), func (0)
       {
       }
   
       info (char const *filename, unsigned linenumber, char const *function)
       : file (filename ? filename : "<anonymous file>"), line (linenumber), func (function ? function : "<anonymous function> ()")
       {
       }
     };
   
     info m_info;
   };
   
   struct exception : std::exception
   {
     virtual const char *what() const throw ()
     {
       return m_what ? m_what : "";
     }
   
   protected:
     void what (srcinf const &si, char const *fmt, va_list ap) throw ()
     {
       char buf[EBUFSIZE];
       char *bufptr = buf;
       int l = snprintf (bufptr, EBUFSIZE - 1, "%s(#%d): %s: ", si.file (), si.line (), si.func ());
       vsnprintf (bufptr + l, EBUFSIZE - 1, fmt, ap);
       what (buf);
     }
   
     void what (srcinf const &si, char const *w) throw ()
     {
       char buf[EBUFSIZE];
       snprintf (buf, EBUFSIZE - 1, "%s(#%d): %s: %s", si.file (), si.line (), si.func (), w);
       what (buf);
     }
   
     void what (char const *fmt, va_list ap) throw ()
     {
       char buf[EBUFSIZE];
       vsnprintf (buf, EBUFSIZE - 1, fmt, ap);
       what (buf);
     }
   
     void what (char const *w) throw ()
     {
       std::copy (w, w + EBUFSIZE, m_what);
     }
   
     exception ()
     {
     }
   
   private:
     char m_what[EBUFSIZE];
   };
} // namespace impl

#define ADD_EXCEPTION(name, base)                                               \
struct name ## _exception : base                                                \
{                                                                               \
  explicit name ## _exception ()                                                \
  {                                                                             \
    what (#name " exception thrown");                                           \
  }                                                                             \
                                                                                \
  explicit name ## _exception (impl::srcinf const &si)                          \
  {                                                                             \
    what (si, #name " exception thrown");                                       \
  }                                                                             \
                                                                                \
  explicit name ## _exception (char const *fmt, ...)                            \
  {                                                                             \
    va_list ap;                                                                 \
    va_start (ap, fmt);                                                         \
    what (fmt, ap);                                                             \
    va_end (ap);                                                                \
  }                                                                             \
                                                                                \
  explicit name ## _exception (impl::srcinf const &si, char const *fmt, ...)    \
  {                                                                             \
    va_list ap;                                                                 \
    va_start (ap, fmt);                                                         \
    what (si, fmt, ap);                                                         \
    va_end (ap);                                                                \
  }                                                                             \
}

ADD_EXCEPTION (nullpointer, impl::exception);
ADD_EXCEPTION (balloc, impl::exception);
ADD_EXCEPTION (module, impl::exception);
ADD_EXCEPTION (factory, impl::exception);

#endif // ERMYTH_EXCEPTION_H
