/**
 * voice.C: This file contains code for the ChanServ VOICE functions.
 *
 * Copyright © 2007 Pippijn van Steenhoven / The Ermyth Team
 * Rights to this code are as documented in COPYING.
 *
 *
 * Portions of this file were derived from sources bearing the following license:
 * Copyright © 2005 William Pitcock, et al.
 * Rights to this code are as documented in doc/pod/license.pod.
 *
 * $Id: voice.C,v 1.8 2007/09/22 14:27:27 pippijn Exp $
 */

#include "atheme.h"
#include <ermyth/module.h>
#include <account/mychan.h>
#include <account/chanacs.h>

static char const rcsid[] = "$Id: voice.C,v 1.8 2007/09/22 14:27:27 pippijn Exp $";

REGISTER_MODULE ("chanserv/voice", false, "The Ermyth Team <http://ermyth.xinutec.org>");

static void cs_cmd_voice (sourceinfo_t *si, int parc, char *parv[]);
static void cs_cmd_devoice (sourceinfo_t *si, int parc, char *parv[]);

command_t const cs_voice = { "VOICE", N_("Gives channel voice to a user."), AC_NONE, 2, cs_cmd_voice };
command_t const cs_devoice = { "DEVOICE", N_("Removes channel voice from a user."), AC_NONE, 2, cs_cmd_devoice };

E cmdvec cs_cmdtree;
E helpvec cs_helptree;

bool
_modinit (module *m)
{
  cs_cmdtree << cs_voice;
  cs_cmdtree << cs_devoice;

  help_addentry (cs_helptree, "VOICE", "help/chanserv/voice", NULL);
  help_addentry (cs_helptree, "DEVOICE", "help/chanserv/devoice", NULL);

  return true;
}

void
_moddeinit ()
{
  cs_cmdtree >> cs_voice;
  cs_cmdtree >> cs_devoice;

  help_delentry (cs_helptree, "VOICE");
  help_delentry (cs_helptree, "DEVOICE");
}

static void
cs_cmd_voice (sourceinfo_t *si, int parc, char *parv[])
{
  char *chan = parv[0];
  char *nick = parv[1];
  mychan_t *mc;
  user_t *tu;
  chanuser_t *cu;

  if (!chan)
    {
      command_fail (si, fault::needmoreparams, STR_INSUFFICIENT_PARAMS, "VOICE");
      command_fail (si, fault::needmoreparams, _("Syntax: VOICE <#channel> [nickname]"));
      return;
    }

  mc = mychan_t::find (chan);
  if (!mc)
    {
      command_fail (si, fault::nosuch_target, _("\2%s\2 is not registered."), chan);
      return;
    }

  if (mc->find_metadata ("private:close:closer"))
    {
      command_fail (si, fault::noprivs, _("\2%s\2 is closed."), chan);
      return;
    }

  if (!chanacs_source_has_flag (mc, si, CA_VOICE))
    {
      command_fail (si, fault::noprivs, _("You are not authorized to perform this operation."));
      return;
    }

  /* figure out who we're going to voice */
  if (!nick)
    tu = si->su;
  else
    {
      if (!(tu = user_find_named (nick)))
	{
	  command_fail (si, fault::nosuch_target, _("\2%s\2 is not online."), nick);
	  return;
	}
    }

  if (is_internal_client (tu))
    return;

  cu = chanuser_find (mc->chan, tu);
  if (!cu)
    {
      command_fail (si, fault::nosuch_target, _("\2%s\2 is not on \2%s\2."), tu->nick, mc->name);
      return;
    }

  modestack_mode_param (chansvs.nick, mc->chan, MTYPE_ADD, 'v', CLIENT_NAME (tu));
  cu->modes |= CMODE_VOICE;

  if (si->c == NULL && tu != si->su)
    notice (chansvs.nick, tu->nick, "You have been voiced on %s by %s", mc->name, get_source_name (si));

  logcommand (si, CMDLOG_SET, "%s VOICE %s!%s@%s", mc->name, tu->nick, tu->user, tu->vhost);
  if (!chanuser_find (mc->chan, si->su))
    command_success_nodata (si, _("\2%s\2 has been voiced on \2%s\2."), tu->nick, mc->name);
}

static void
cs_cmd_devoice (sourceinfo_t *si, int parc, char *parv[])
{
  char *chan = parv[0];
  char *nick = parv[1];
  mychan_t *mc;
  user_t *tu;
  chanuser_t *cu;

  if (!chan)
    {
      command_fail (si, fault::needmoreparams, STR_INSUFFICIENT_PARAMS, "DEVOICE");
      command_fail (si, fault::needmoreparams, _("Syntax: DEVOICE <#channel> [nickname]"));
      return;
    }

  mc = mychan_t::find (chan);
  if (!mc)
    {
      command_fail (si, fault::nosuch_target, _("\2%s\2 is not registered."), chan);
      return;
    }

  if (!chanacs_source_has_flag (mc, si, CA_VOICE))
    {
      command_fail (si, fault::noprivs, _("You are not authorized to perform this operation."));
      return;
    }

  /* figure out who we're going to devoice */
  if (!nick)
    tu = si->su;
  else
    {
      if (!(tu = user_find_named (nick)))
	{
	  command_fail (si, fault::nosuch_target, _("\2%s\2 is not online."), nick);
	  return;
	}
    }

  if (is_internal_client (tu))
    return;

  cu = chanuser_find (mc->chan, tu);
  if (!cu)
    {
      command_fail (si, fault::nosuch_target, _("\2%s\2 is not on \2%s\2."), tu->nick, mc->name);
      return;
    }

  modestack_mode_param (chansvs.nick, mc->chan, MTYPE_DEL, 'v', CLIENT_NAME (tu));
  cu->modes &= ~CMODE_VOICE;

  if (si->c == NULL && tu != si->su)
    notice (chansvs.nick, tu->nick, "You have been devoiced on %s by %s", mc->name, get_source_name (si));

  logcommand (si, CMDLOG_SET, "%s DEVOICE %s!%s@%s", mc->name, tu->nick, tu->user, tu->vhost);
  if (!chanuser_find (mc->chan, si->su))
    command_success_nodata (si, _("\2%s\2 has been devoiced on \2%s\2."), tu->nick, mc->name);
}

/* vim:cinoptions=>s,e0,n0,f0,{0,}0,^0,=s,ps,t0,c3,+s,(2s,us,)20,*30,gs,hs
 * vim:ts=8
 * vim:sw=8
 * vim:noexpandtab
 */
