/**
 * freeze.C: Gives services the ability to freeze nicknames
 *
 * Copyright © 2007 Pippijn van Steenhoven / The Ermyth Team
 * Rights to this code are as documented in COPYING.
 *
 *
 * Portions of this file were derived from sources bearing the following license:
 * Copyright © 2005-2007 Patrick Fish, et al.
 * Rights to this code are as documented in doc/pod/license.pod.
 *
 * $Id: freeze.C,v 1.8 2007/09/22 14:27:27 pippijn Exp $
 */

#include "atheme.h"
#include <util/numeric.h>
#include <libermyth.h>
#include <ermyth/module.h>
#include <account/myuser.h>

static char const rcsid[] = "$Id: freeze.C,v 1.8 2007/09/22 14:27:27 pippijn Exp $";

REGISTER_MODULE ("nickserv/freeze", false, "The Ermyth Team <http://ermyth.xinutec.org>");

static void ns_cmd_freeze (sourceinfo_t *si, int parc, char *parv[]);

/* FREEZE ON|OFF -- don't pollute the root with THAW */
command_t const ns_freeze = { "FREEZE", N_("Freezes an account."), PRIV_USER_ADMIN, 3, ns_cmd_freeze };

E cmdvec ns_cmdtree;
E helpvec ns_helptree;

bool
_modinit (module *m)
{
  ns_cmdtree << ns_freeze;
  help_addentry (ns_helptree, "FREEZE", "help/nickserv/freeze", NULL);

  return true;
}

void
_moddeinit ()
{
  ns_cmdtree >> ns_freeze;
  help_delentry (ns_helptree, "FREEZE");
}

static void
ns_cmd_freeze (sourceinfo_t *si, int parc, char *parv[])
{
  myuser_t *mu;
  char *target = parv[0];
  char *action = parv[1];
  char *reason = parv[2];
  user_t *u;

  if (!target || !action)
    {
      command_fail (si, fault::needmoreparams, STR_INSUFFICIENT_PARAMS, "FREEZE");
      command_fail (si, fault::needmoreparams, _("Usage: FREEZE <nickname> <ON|OFF> [reason]"));
      return;
    }

  mu = myuser_t::find_ext (target);

  if (!mu)
    {
      command_fail (si, fault::nosuch_target, _("\2%s\2 is not registered."), target);
      return;
    }

  if (!strcasecmp (action, "ON"))
    {
      if (!reason)
	{
	  command_fail (si, fault::needmoreparams, STR_INSUFFICIENT_PARAMS, "FREEZE");
	  command_fail (si, fault::needmoreparams, _("Usage: FREEZE <nickname> ON <reason>"));
	  return;
	}

      if (is_soper (mu))
	{
	  command_fail (si, fault::badparams, _("The account \2%s\2 belongs to a services operator; it cannot be frozen."), target);
	  return;
	}

      if (mu->find_metadata ("private:freeze:freezer"))
	{
	  command_fail (si, fault::badparams, _("\2%s\2 is already frozen."), target);
	  return;
	}

      mu->add_metadata ("private:freeze:freezer", get_oper_name (si));
      mu->add_metadata ("private:freeze:reason", reason);
      mu->add_metadata ("private:freeze:timestamp", itoa (NOW));
      /* log them out */
      while (!mu->logins.empty ())
        {
          u = mu->logins.back ();
          if (!phandler->ircd_on_logout (u->nick, mu->name, NULL))
            {
              u->myuser = NULL;
              mu->logins.erase (u);
            }
        }
      mu->flags |= MU_NOBURSTLOGIN;

      wallops ("%s froze the account \2%s\2 (%s).", get_oper_name (si), target, reason);
      logcommand (si, CMDLOG_ADMIN, "FREEZE %s ON", target);
      command_success_nodata (si, _("\2%s\2 is now frozen."), target);
    }
  else if (!strcasecmp (action, "OFF"))
    {
      if (!mu->find_metadata ("private:freeze:freezer"))
	{
	  command_fail (si, fault::badparams, _("\2%s\2 is not frozen."), target);
	  return;
	}

      mu->del_metadata ("private:freeze:freezer");
      mu->del_metadata ("private:freeze:reason");
      mu->del_metadata ("private:freeze:timestamp");

      wallops ("%s thawed the account \2%s\2.", get_oper_name (si), target);
      logcommand (si, CMDLOG_ADMIN, "FREEZE %s OFF", target);
      command_success_nodata (si, _("\2%s\2 has been thawed"), target);
    }
  else
    {
      command_fail (si, fault::needmoreparams, STR_INSUFFICIENT_PARAMS, "FREEZE");
      command_fail (si, fault::needmoreparams, _("Usage: FREEZE <account> <ON|OFF> [reason]"));
    }
}
