/**
 * register.C: This file contains code for the NickServ REGISTER function.
 *
 * Copyright © 2007 Pippijn van Steenhoven / The Ermyth Team
 * Rights to this code are as documented in COPYING.
 *
 *
 * Portions of this file were derived from sources bearing the following license:
 * Copyright © 2005 William Pitcock, et al.
 * Rights to this code are as documented in doc/pod/license.pod.
 *
 * $Id: register.C,v 1.11 2007/09/22 14:27:28 pippijn Exp $
 */

#include "atheme.h"
#include <util/numeric.h>
#include <libermyth.h>
#include <ermyth/module.h>
#include <account/mynick.h>
#include <account/myuser.h>
#include <boost/foreach.hpp>

static char const rcsid[] = "$Id: register.C,v 1.11 2007/09/22 14:27:28 pippijn Exp $";

REGISTER_MODULE ("nickserv/register", false, "The Ermyth Team <http://ermyth.xinutec.org>");

static void ns_cmd_register (sourceinfo_t *si, int parc, char *parv[]);

command_t const ns_register = { "REGISTER", N_("Registers a nickname."), AC_NONE, 3, ns_cmd_register };

E cmdvec ns_cmdtree;
E helpvec ns_helptree;

static unsigned tcnt = 0;

static void
ns_cmd_register (sourceinfo_t *si, int parc, char *parv[])
{
  myuser_t *mu;
  myuser_t *tmu;
  mynick_t *mn;
  char *account;
  char *pass = parv[0];
  char *email = parv[1];
  char lau[BUFSIZE], lao[BUFSIZE];

  if (si->smu)
    {
      command_fail (si, fault::already_authed, _("You are already logged in."));
      return;
    }

  if (nicksvs.no_nick_ownership || si->su == NULL)
    account = parv[0], pass = parv[1], email = parv[2];
  else
    account = si->su->nick, pass = parv[0], email = parv[1];

  if (!account || !pass || !email)
    {
      command_fail (si, fault::needmoreparams, STR_INSUFFICIENT_PARAMS, "REGISTER");
      if (nicksvs.no_nick_ownership || si->su == NULL)
	command_fail (si, fault::needmoreparams, _("Syntax: REGISTER <account> <password> <email>"));
      else
	command_fail (si, fault::needmoreparams, _("Syntax: REGISTER <password> <email>"));
      return;
    }

  if ((strlen (pass) > 32) || (strlen (email) >= EMAILLEN))
    {
      command_fail (si, fault::badparams, STR_INVALID_PARAMS, "REGISTER");
      return;
    }

  if (!nicksvs.no_nick_ownership && si->su == NULL && user_find_named (account))
    {
      command_fail (si, fault::noprivs, _("A user matching this account is already on IRC."));
      return;
    }

  if (!nicksvs.no_nick_ownership && IsDigit (*account))
    {
      command_fail (si, fault::badparams, _("For security reasons, you can't register your UID."));
      command_fail (si, fault::badparams, _("Please change to a real nickname, and try again."));
      return;
    }

  if (nicksvs.no_nick_ownership || si->su == NULL)
    {
      if (strchr (account, ' ') || strchr (account, '\n') || strchr (account, '\r') || account[0] == '=' || account[0] == '#' || account[0] == '@' || account[0] == '+' || account[0] == '%' || account[0] == '!' || strchr (account, ','))
	{
	  command_fail (si, fault::badparams, _("The account name \2%s\2 is invalid."), account);
	  return;
	}
    }

  if ((si->su != NULL && !strcasecmp (pass, si->su->nick)) || !strcasecmp (pass, account))
    {
      command_fail (si, fault::badparams, _("You cannot use your nickname as a password."));
      if (nicksvs.no_nick_ownership || si->su == NULL)
	command_fail (si, fault::needmoreparams, _("Syntax: REGISTER <account> <password> <email>"));
      else
	command_fail (si, fault::needmoreparams, _("Syntax: REGISTER <password> <email>"));
      return;
    }

  if (!validemail (email))
    {
      command_fail (si, fault::badparams, _("\2%s\2 is not a valid email address."), email);
      return;
    }

  /* make sure it isn't registered already */
  if (nicksvs.no_nick_ownership ? myuser_t::find (account) != NULL : mynick_t::find (account) != NULL)
    {
      command_fail (si, fault::alreadyexists, _("\2%s\2 is already registered."), account);
      return;
    }

  if (!user_t::callback.can_register (si, account, email))
    return;

  /* make sure they're within limits */
  tcnt = 0;
  foreach (myuser_t::pair_type &mupair, myuser_t::map)
    {
      tmu = mupair.second;
      
      if (!strcasecmp (email, tmu->email))
        tcnt++;
      
      /* optimization: if tcnt >= me.maxusers, quit iterating. -nenolod */
      if (tcnt >= me.maxusers)
        break;
    }
  
  if (tcnt >= me.maxusers)
    {
      command_fail (si, fault::toomany, _("\2%s\2 has too many nicknames registered."), email);
      return;
    }

  mu = myuser_t::create (account, pass, email, config_options.defuflags | MU_NOBURSTLOGIN);
  mu->registered = NOW;
  mu->lastlogin = NOW;
  if (!nicksvs.no_nick_ownership)
    {
      mn = mynick_t::create (mu, mu->name);
      mn->registered = NOW;
      mn->lastseen = NOW;
    }

  if (me.auth == AUTH_EMAIL)
    {
      char *key = gen_pw (12);
      mu->flags |= MU_WAITAUTH;

      mu->add_metadata ("private:verify:register:key", key);
      mu->add_metadata ("private:verify:register:timestamp", itoa (NOW));

      if (!sendemail (si->su != NULL ? si->su : si->service->me, EMAIL_REGISTER, mu, key))
	{
	  command_fail (si, fault::emailfail, _("Sending email failed, sorry! Registration aborted."));
          mu->refcnt_dec ();
	  sfree (key);
	  return;
	}

      command_success_nodata (si, _("An email containing nickname activation instructions has been sent to \2%s\2."), mu->email);
      command_success_nodata (si, _("If you do not complete registration within one day, your nickname will expire."));

      sfree (key);
    }

  if (si->su != NULL)
    {
      si->su->myuser = mu;
      mu->logins.insert (si->su);

      if (!(mu->flags & MU_WAITAUTH))
	/* only grant ircd registered status if it's verified */
	phandler->ircd_on_login (si->su->nick, mu->name, NULL);
    }

  if (!nicksvs.no_nick_ownership && si->su != NULL)
    snoop ("REGISTER: \2%s\2 to \2%s\2", account, email);
  else
    snoop ("REGISTER: \2%s\2 to \2%s\2 by \2%s\2", account, email, si->su != NULL ? si->su->nick : get_source_name (si));
  logcommand (si, CMDLOG_REGISTER, "REGISTER to %s", email);
  if (is_soper (mu))
    {
      wallops ("%s registered the nick \2%s\2 and gained services operator privileges.", get_oper_name (si), mu->name);
      snoop ("SOPER: \2%s\2 as \2%s\2", get_oper_name (si), mu->name);
    }

  command_success_nodata (si, _("\2%s\2 is now registered to \2%s\2."), mu->name, mu->email);
  command_success_nodata (si, _("The password is \2%s\2. Please write this down for future reference."), pass);
  mu->callback.registered (mu);

  if (si->su != NULL)
    {
      snprintf (lau, BUFSIZE, "%s@%s", si->su->user, si->su->vhost);
      mu->add_metadata ("private:host:vhost", lau);

      snprintf (lao, BUFSIZE, "%s@%s", si->su->user, si->su->host);
      mu->add_metadata ("private:host:actual", lao);
    }
}

bool
_modinit (module *m)
{
  ns_cmdtree << ns_register;
  help_addentry (ns_helptree, "REGISTER", "help/nickserv/register", NULL);

  return true;
}

void
_moddeinit ()
{
  ns_cmdtree >> ns_register;
  help_delentry (ns_helptree, "REGISTER");
}
