/**
 * soper.C: Dynamic services operator privileges
 *
 * Copyright © 2007 Pippijn van Steenhoven / The Ermyth Team
 * Rights to this code are as documented in COPYING.
 *
 *
 * Portions of this file were derived from sources bearing the following license:
 * Copyright © 2006 Atheme Development Group
 * Rights to this code are as documented in doc/pod/license.pod.
 *
 * $Id: soper.C,v 1.9 2007/09/22 14:27:28 pippijn Exp $
 */

#include "atheme.h"
#include <ermyth/module.h>
#include <account/myuser.h>

static char const rcsid[] = "$Id: soper.C,v 1.9 2007/09/22 14:27:28 pippijn Exp $";

REGISTER_MODULE ("operserv/soper", false, "The Ermyth Team <http://ermyth.xinutec.org>");

static void os_cmd_soper (sourceinfo_t *si, int parc, char *parv[]);
static void os_cmd_soper_list (sourceinfo_t *si, int parc, char *parv[]);
static void os_cmd_soper_listclass (sourceinfo_t *si, int parc, char *parv[]);
static void os_cmd_soper_add (sourceinfo_t *si, int parc, char *parv[]);
static void os_cmd_soper_del (sourceinfo_t *si, int parc, char *parv[]);

command_t const os_soper = { "SOPER", N_("Shows and changes services operator privileges."), AC_NONE, 3, os_cmd_soper };

command_t const os_soper_list = { "LIST", N_("Lists services operators."), PRIV_VIEWPRIVS, 0, os_cmd_soper_list };
command_t const os_soper_listclass = { "LISTCLASS", N_("Lists operclasses."), PRIV_VIEWPRIVS, 0, os_cmd_soper_listclass };
command_t const os_soper_add = { "ADD", N_("Grants services operator privileges to an account."), PRIV_GRANT, 2, os_cmd_soper_add };
command_t const os_soper_del = { "DEL", N_("Removes services operator privileges from an account."), PRIV_GRANT, 1, os_cmd_soper_del };

E cmdvec os_cmdtree;
E helpvec os_helptree;
cmdvec os_soper_cmds;

bool
_modinit (module *m)
{
  os_cmdtree << os_soper;
  os_soper_cmds << os_soper_list;
  os_soper_cmds << os_soper_listclass;
  os_soper_cmds << os_soper_add;
  os_soper_cmds << os_soper_del;

  help_addentry (os_helptree, "SOPER", "help/operserv/soper", NULL);

  return true;
}

void
_moddeinit ()
{
  os_cmdtree >> os_soper;
  os_soper_cmds >> os_soper_list;
  os_soper_cmds >> os_soper_listclass;
  os_soper_cmds >> os_soper_add;
  os_soper_cmds >> os_soper_del;

  help_delentry (os_helptree, "SOPER");
}

static void
os_cmd_soper (sourceinfo_t *si, int parc, char *parv[])
{
  command_t const *c;

  if (!has_any_privs (si))
    {
      command_fail (si, fault::noprivs, _("You are not authorized to perform this operation."));
      return;
    }

  if (parc < 1)
    {
      command_fail (si, fault::needmoreparams, STR_INSUFFICIENT_PARAMS, "SOPER");
      command_fail (si, fault::needmoreparams, _("Syntax: SOPER LIST|LISTCLASS|ADD|DEL [account] [operclass]"));
      return;
    }

  c = os_soper_cmds.find (parv[0]);
  if (c == NULL)
    {
      command_fail (si, fault::badparams, _("Invalid command. Use \2/%s%s help\2 for a command listing."), (ircd->uses_rcommand == false) ? "msg " : "", si->service->disp);
      return;
    }

  c->exec (si->service, si, parc - 1, parv + 1);
}

static void
os_cmd_soper_list (sourceinfo_t *si, int parc, char *parv[])
{
  char const *typestr;

  logcommand (si, CMDLOG_GET, "SOPER LIST");
  command_success_nodata (si, "%-20s %-5s %-20s", _("Account"), _("Type"), _("Operclass"));
  command_success_nodata (si, "%-20s %-5s %-20s", "--------------------", "-----", "--------------------");
  foreach (soper_t *soper, soper_t::list)
    {
      if (!(soper->flags & SOPER_CONF))
        typestr = "DB";
      else if (soper->myuser != NULL)
        typestr = "Conf";
      else
        typestr = "Conf*";
      command_success_nodata (si, "%-20s %-5s %-20s", soper->myuser != NULL ? soper->myuser->name : soper->name, typestr, soper->classname);
    }
  command_success_nodata (si, "%-20s %-5s %-20s", "--------------------", "-----", "--------------------");
  command_success_nodata (si, _("End of services operator list"));
}

static void
os_cmd_soper_listclass (sourceinfo_t *si, int parc, char *parv[])
{
  logcommand (si, CMDLOG_GET, "SOPER LISTCLASS");
  command_success_nodata (si, _("Oper class list:"));

  foreach (operclass_t *oc, operclass_t::list)
    command_success_nodata (si, "%c%c %s", has_all_operclass (si, oc) ? '+' : '-', oc->flags & OPERCLASS_NEEDOPER ? '*' : ' ', oc->name);

  command_success_nodata (si, _("End of oper class list"));
}

static void
os_cmd_soper_add (sourceinfo_t *si, int parc, char *parv[])
{
  myuser_t *mu;
  operclass_t *operclass;

  if (parc < 2)
    {
      command_fail (si, fault::needmoreparams, STR_INSUFFICIENT_PARAMS, "SOPER ADD");
      command_fail (si, fault::needmoreparams, _("Syntax: SOPER ADD <account> <operclass>"));
      return;
    }

  mu = myuser_t::find_ext (parv[0]);
  if (mu == NULL)
    {
      command_fail (si, fault::nosuch_target, _("\2%s\2 is not registered."), parv[0]);
      return;
    }

  if (is_conf_soper (mu))
    {
      command_fail (si, fault::noprivs, _("You may not modify \2%s\2's operclass as it is defined in the configuration file."), mu->name);
      return;
    }

  operclass = operclass_find (parv[1]);
  if (operclass == NULL)
    {
      command_fail (si, fault::nosuch_target, _("No such oper class \2%s\2."), parv[1]);
      return;
    }
  else if (mu->soper != NULL && mu->soper->operclass == operclass)
    {
      command_fail (si, fault::nochange, _("Oper class for \2%s\2 is already set to \2%s\2."), mu->name, operclass->name);
      return;
    }

  if (!has_all_operclass (si, operclass))
    {
      command_fail (si, fault::noprivs, _("Oper class \2%s\2 has more privileges than you."), operclass->name);
      return;
    }
  else if (mu->soper != NULL && mu->soper->operclass != NULL && !has_all_operclass (si, mu->soper->operclass))
    {
      command_fail (si, fault::noprivs, _("Oper class for \2%s\2 is set to \2%s\2 which you are not authorized to change."), mu->name, mu->soper->operclass->name);
      return;
    }

  wallops ("\2%s\2 is changing oper class for \2%s\2 to \2%s\2", get_oper_name (si), mu->name, operclass->name);
  snoop ("SOPER:ADD: \2%s\2 to \2%s\2 by \2%s\2", mu->name, operclass->name, get_oper_name (si));
  logcommand (si, CMDLOG_ADMIN, "SOPER ADD %s %s", mu->name, operclass->name);
  if (is_soper (mu))
    soper_delete (mu->soper);
  soper_add (mu->name, operclass->name, 0);
  command_success_nodata (si, _("Set class for \2%s\2 to \2%s\2."), mu->name, operclass->name);
}

static void
os_cmd_soper_del (sourceinfo_t *si, int parc, char *parv[])
{
  myuser_t *mu;

  if (parc < 1)
    {
      command_fail (si, fault::needmoreparams, STR_INSUFFICIENT_PARAMS, "SOPER DEL");
      command_fail (si, fault::needmoreparams, _("Syntax: SOPER DEL <account>"));
      return;
    }

  mu = myuser_t::find_ext (parv[0]);
  if (mu == NULL)
    {
      command_fail (si, fault::nosuch_target, _("\2%s\2 is not registered."), parv[0]);
      return;
    }

  if (is_conf_soper (mu))
    {
      command_fail (si, fault::noprivs, _("You may not modify \2%s\2's operclass as it is defined in the configuration file."), mu->name);
      return;
    }

  if (!is_soper (mu))
    {
      command_fail (si, fault::nochange, _("\2%s\2 does not have an operclass set."), mu->name);
      return;
    }
  else if (mu->soper->operclass != NULL && !has_all_operclass (si, mu->soper->operclass))
    {
      command_fail (si, fault::noprivs, _("Oper class for \2%s\2 is set to \2%s\2 which you are not authorized to change."), mu->name, mu->soper->operclass->name);
      return;
    }

  wallops ("\2%s\2 is removing oper class for \2%s\2", get_oper_name (si), mu->name);
  snoop ("SOPER:DELETE: \2%s\2 by \2%s\2", mu->name, get_oper_name (si));
  logcommand (si, CMDLOG_ADMIN, "SOPER DEL %s", mu->name);
  soper_delete (mu->soper);
  command_success_nodata (si, _("Removed class for \2%s\2."), mu->name);
}
