/**
 * dreamforge.C: This file contains protocol support for bahamut-based ircd.
 *
 * Copyright © 2007 Pippijn van Steenhoven / The Ermyth Team
 * Rights to this code are as documented in COPYING.
 *
 *
 * Portions of this file were derived from sources bearing the following license:
 * Copyright © 2003-2004 E. Will et al.
 * Copyright © 2005-2006 Atheme Development Group
 * Rights to this code are documented in doc/pod/license.pod.
 *
 * $Id: dreamforge.C,v 1.8 2007/09/22 14:27:28 pippijn Exp $
 */

#include <boost/foreach.hpp>

#include "atheme.h"
#include <util/time.h>
#include <libermyth.h>
#include "servers.h"
#include <account/mynick.h>
#include "uplink.h"
#include "pmodule.h"
#include "protocol/dreamforge.h"

static char const rcsid[] = "$Id: dreamforge.C,v 1.8 2007/09/22 14:27:28 pippijn Exp $";

/* *INDENT-OFF* */

namespace protocol
{
  static ircd_t DreamForge = {
    "DreamForge 4.6.7 and later", /* IRCd name */
    "$",                          /* TLD Prefix, used by Global. */
    false,                        /* Whether or not we use IRCNet/TS6 UID */
    false,                        /* Whether or not we use RCOMMAND */
    false,                        /* Whether or not we support channel owners. */
    false,                        /* Whether or not we support channel protection. */
    false,                        /* Whether or not we support halfops. */
    false,                        /* Whether or not we use P10 */
    false,                        /* Whether or not we use vHosts. */
    0,                            /* Oper-only cmodes */
    0,                            /* Integer flag for owner channel flag. */
    0,                            /* Integer flag for protect channel flag. */
    0,                            /* Integer flag for halfops. */
    "+",                          /* Mode we set for owner. */
    "+",                          /* Mode we set for protect. */
    "+",                          /* Mode we set for halfops. */
    PROTOCOL_DREAMFORGE,          /* Protocol type */
    0,                            /* Permanent cmodes */
    "b",                          /* Ban-like cmodes */
    0,                            /* Except mchar */
    0,                            /* Invex mchar */
    0                             /* Flags */
  };
  
  struct dreamforge_handler : handler
  {
    dreamforge_handler ();
    virtual ~dreamforge_handler ();
  
    virtual unsigned int server_login (void);
    virtual void introduce_nick (user_t *u);
    virtual void invite_sts (user_t *source, user_t *target, channel_t *channel);
    virtual void quit_sts (user_t *u, char const * const reason);
    virtual void wallops_sts (char const * const text);
    virtual void join_sts (channel_t *c, user_t *u, bool isnew, char const * const modes);
    virtual void kick (char const * const from, char const * const channel, char const * const to, char const * const reason);
    virtual void privmsg (char const * const from, char const * const target, char const * const fmt, ...);
    virtual void notice_user_sts (user_t *from, user_t *target, char const * const text);
    virtual void notice_global_sts (user_t *from, char const * const mask, char const * const text);
    virtual void notice_channel_sts (user_t *from, channel_t *target, char const * const text);
    virtual void numeric_sts (char const * const from, int numeric, char const * const target, char const * const fmt, ...);
    virtual void skill (char const * const from, char const * const nick, char const * const fmt, ...);
    virtual void part_sts (channel_t *c, user_t *u);
    virtual void kline_sts (char const * const server, char const * const user, char const * const host, long duration, char const * const reason);
    virtual void unkline_sts (char const * const server, char const * const user, char const * const host);
    virtual void topic_sts (channel_t *c, char const * const setter, time_t ts, time_t prevts, char const * const topic);
    virtual void mode_sts (char const * const sender, channel_t *target, char const * const modes);
    virtual void ping_sts (void);
    virtual void ircd_on_login (char const * const origin, char const * const user, char const * const wantedhost);
    virtual bool ircd_on_logout (char const * const origin, char const * const user, char const * const wantedhost);
    virtual void jupe (char const * const server, char const * const reason);
  };
  
  static cmode_t dreamforge_mode_list[] = {
    { 'i', CMODE_INVITE   },
    { 'm', CMODE_MOD      },
    { 'n', CMODE_NOEXT    },
    { 'p', CMODE_PRIV     },
    { 's', CMODE_SEC      },
    { 't', CMODE_TOPIC    },
    { 'c', CMODE_NOCOLOR  },
    { 'R', CMODE_REGONLY  },
    { '\0', 0 }
  };
  
  static extmode_t dreamforge_ignore_mode_list[] = {
    { '\0', 0 }
  };
  
  static cmode_t dreamforge_status_mode_list[] = {
    { 'o', CMODE_OP      },
    { 'v', CMODE_VOICE   },
    { '\0', 0 }
  };
  
  static cmode_t dreamforge_prefix_mode_list[] = {
    { '@', CMODE_OP      },
    { '+', CMODE_VOICE   },
    { '\0', 0 }
  };
  
  /* *INDENT-ON* */
  
  /* login to our uplink */
  unsigned int
  dreamforge_handler::server_login (void)
  {
    int ret;
  
    ret = sts ("PASS %s", curr_uplink->pass);
    if (ret == 1)
      return 1;
  
    me.bursting = true;
  
    sts ("PROTOCTL NOQUIT WATCH=128 SAFELIST");
    sts ("SERVER %s 1 :%s", me.name, me.desc);
  
    services_init ();
  
    return 0;
  }
  
  /* introduce a client */
  void
  dreamforge_handler::introduce_nick (user_t *u)
  {
    sts ("NICK %s 1 %ld %s %s %s 0 :%s", u->nick, u->ts, u->user, u->host, me.name, u->gecos);
    sts (":%s MODE %s +%s", u->nick, u->nick, "io");
  }
  
  /* invite a user to a channel */
  void
  dreamforge_handler::invite_sts (user_t *sender, user_t *target, channel_t *channel)
  {
    sts (":%s INVITE %s %s", sender->nick, target->nick, channel->name);
  }
  
  void
  dreamforge_handler::quit_sts (user_t *u, char const * const reason)
  {
    if (!me.connected)
      return;
  
    sts (":%s QUIT :%s", u->nick, reason);
  }
  
  /* WALLOPS wrapper */
  void
  dreamforge_handler::wallops_sts (char const * const text)
  {
    sts (":%s GLOBOPS :%s", me.name, text);
  }
  
  /* join a channel */
  void
  dreamforge_handler::join_sts (channel_t *c, user_t *u, bool isnew, char const * const modes)
  {
    if (isnew)
      {
        sts (":%s JOIN %s", u->nick, c->name);
        if (modes[0] && modes[1])
          sts (":%s MODE %s %s", u->nick, c->name, modes);
      }
    else
      {
        sts (":%s JOIN %s", u->nick, c->name);
      }
    sts (":%s MODE %s +o %s %ld", u->nick, c->name, u->nick, c->ts);
  }
  
  /* kicks a user from a channel */
  void
  dreamforge_handler::kick (char const * const from, char const * const channel, char const * const to, char const * const reason)
  {
    channel_t *chan = channel_find (channel);
    user_t *user = user_find (to);
  
    if (!chan || !user)
      return;
  
    sts (":%s KICK %s %s :%s", from, channel, to, reason);
  
    chanuser_delete (chan, user);
  }
  
  /* PRIVMSG wrapper */
  void
  dreamforge_handler::privmsg (char const * const from, char const * const target, char const * const fmt, ...)
  {
    va_list ap;
    char buf[BUFSIZE];
  
    va_start (ap, fmt);
    vsnprintf (buf, BUFSIZE, fmt, ap);
    va_end (ap);
  
    sts (":%s PRIVMSG %s :%s", from, target, buf);
  }
  
  /* NOTICE wrapper */
  void
  dreamforge_handler::notice_user_sts (user_t *from, user_t *target, char const * const text)
  {
    sts (":%s NOTICE %s :%s", from ? from->nick : me.name, target->nick, text);
  }
  
  void
  dreamforge_handler::notice_global_sts (user_t *from, char const * const mask, char const * const text)
  {
    if (!strcmp (mask, "*"))
      foreach (tld_t *tld, tld_t::list)
        sts (":%s NOTICE %s*%s :%s", from ? from->nick : me.name, ircd->tldprefix, tld->name, text);
    else
      sts (":%s NOTICE %s%s :%s", from ? from->nick : me.name, ircd->tldprefix, mask, text);
  }
  
  void
  dreamforge_handler::notice_channel_sts (user_t *from, channel_t *target, char const * const text)
  {
    sts (":%s NOTICE %s :%s", from ? from->nick : me.name, target->name, text);
  }
  
  void
  dreamforge_handler::numeric_sts (char const * const from, int numeric, char const * const target, char const * const fmt, ...)
  {
    va_list ap;
    char buf[BUFSIZE];
  
    va_start (ap, fmt);
    vsnprintf (buf, BUFSIZE, fmt, ap);
    va_end (ap);
  
    sts (":%s %d %s %s", from, numeric, target, buf);
  }
  
  /* KILL wrapper */
  void
  dreamforge_handler::skill (char const * const from, char const * const nick, char const * const fmt, ...)
  {
    va_list ap;
    char buf[BUFSIZE];
  
    va_start (ap, fmt);
    vsnprintf (buf, BUFSIZE, fmt, ap);
    va_end (ap);
  
    sts (":%s KILL %s :%s!%s!%s (%s)", from, nick, from, from, from, buf);
  }
  
  /* PART wrapper */
  void
  dreamforge_handler::part_sts (channel_t *c, user_t *u)
  {
    sts (":%s PART %s", u->nick, c->name);
  }
  
  /* server-to-server KLINE wrapper */
  void
  dreamforge_handler::kline_sts (char const * const server, char const * const user, char const * const host, long duration, char const * const reason)
  {
    if (!me.connected)
      return;
  
    sts (":%s AKILL %s %s %ld %s %ld :%s", me.name, host, user, duration, opersvs.nick, NOW, reason);
  }
  
  /* server-to-server UNKLINE wrapper */
  void
  dreamforge_handler::unkline_sts (char const * const server, char const * const user, char const * const host)
  {
    if (!me.connected)
      return;
  
    sts (":%s RAKILL %s %s", me.name, host, user);
  }
  
  /* topic wrapper */
  void
  dreamforge_handler::topic_sts (channel_t *c, char const * const setter, time_t ts, time_t prevts, char const * const topic)
  {
    if (!me.connected)
      return;
  
    if (ts < prevts || prevts == 0)
      sts (":%s TOPIC %s %s %ld :%s", chansvs.nick, c->name, setter, ts, topic);
    else if (prevts > 1)
      {
        ts = prevts - 1;
        sts (":%s TOPIC %s %s %ld :%s", chansvs.nick, c->name, "topictime.wrong", ts, topic);
        c->topicts = ts;
      }
    else
      {
        notice (chansvs.nick, c->name, "Unable to change topic to: %s", topic);
        c->topicts = 1;
      }
  }
  
  /* mode wrapper */
  void
  dreamforge_handler::mode_sts (char const * const sender, channel_t *target, char const * const modes)
  {
    if (!me.connected)
      return;
  
    sts (":%s MODE %s %s", sender, target->name, modes);
  }
  
  /* ping wrapper */
  void
  dreamforge_handler::ping_sts (void)
  {
    if (!me.connected)
      return;
  
    sts ("PING :%s", me.name);
  }
  
  /* protocol-specific stuff to do on login */
  void
  dreamforge_handler::ircd_on_login (char const * const origin, char const * const user, char const * const wantedhost)
  {
    user_t *u = user_find (origin);
  
    if (!me.connected || u == NULL)
      return;
  
    /* Can only do this for nickserv, and can only record identified
     * state if logged in to correct nick, sorry -- jilles
     */
    if (should_reg_umode (u))
      sts (":%s SVSMODE %s +rd %ld", nicksvs.nick, origin, NOW);
  }
  
  /* protocol-specific stuff to do on login */
  bool
  dreamforge_handler::ircd_on_logout (char const * const origin, char const * const user, char const * const wantedhost)
  {
    if (!me.connected)
      return false;
  
    if (!nicksvs.no_nick_ownership)
      sts (":%s SVSMODE %s -r+d %ld", nicksvs.nick, origin, NOW);
  
    return false;
  }
  
  void
  dreamforge_handler::jupe (char const * const server, char const * const reason)
  {
    if (!me.connected)
      return;
  
    server_delete (server);
    sts (":%s SQUIT %s :%s", opersvs.nick, server, reason);
    sts (":%s SERVER %s 2 :%s", me.name, server, reason);
  }
  
  static void
  m_topic (sourceinfo_t *si, int parc, char *parv[])
  {
    channel_t *c = channel_find (parv[0]);
  
    if (!c)
      return;
  
    handle_topic_from (si, c, parv[1], atol (parv[2]), parv[3]);
  }
  
  static void
  m_ping (sourceinfo_t *si, int parc, char *parv[])
  {
    /* reply to PING's */
    sts (":%s PONG %s %s", me.name, me.name, parv[0]);
  }
  
  static void
  m_pong (sourceinfo_t *si, int parc, char *parv[])
  {
    server_t *s;
  
    /* someone replied to our PING */
    if (!parv[0])
      return;
    s = server_find (parv[0]);
    if (s == NULL)
      return;
    handle_eob (s);
  
    if (irccasecmp (me.actual, parv[0]))
      return;
  
    me.uplinkpong = NOW;
  
    /* -> :test.projectxero.net PONG test.projectxero.net :shrike.malkier.net */
    if (me.bursting)
      {
  #ifdef HAVE_GETTIMEOFDAY
        e_time (burstime, &burstime);
  
        slog (LG_INFO, "m_pong(): finished synching with uplink (%d %s)", (tv2ms (&burstime) > 1000) ? (tv2ms (&burstime) / 1000) : tv2ms (&burstime), (tv2ms (&burstime) > 1000) ? "s" : "ms");
  
        wallops ("Finished synching to network in %d %s.", (tv2ms (&burstime) > 1000) ? (tv2ms (&burstime) / 1000) : tv2ms (&burstime), (tv2ms (&burstime) > 1000) ? "s" : "ms");
  #else
        slog (LG_INFO, "m_pong(): finished synching with uplink");
        wallops ("Finished synching to network.");
  #endif
  
        me.bursting = false;
      }
  }
  
  static void
  m_privmsg (sourceinfo_t *si, int parc, char *parv[])
  {
    if (parc != 2)
      return;
  
    handle_message (si, parv[0], false, parv[1]);
  }
  
  static void
  m_notice (sourceinfo_t *si, int parc, char *parv[])
  {
    if (parc != 2)
      return;
  
    handle_message (si, parv[0], true, parv[1]);
  }
  
  static void
  m_part (sourceinfo_t *si, int parc, char *parv[])
  {
    int chanc;
    char *chanv[256];
    int i;
  
    chanc = sjtoken (parv[0], ',', chanv);
    for (i = 0; i < chanc; i++)
      {
        slog (LG_DEBUG, "m_part(): user left channel: %s -> %s", si->su->nick, chanv[i]);
  
        chanuser_delete (channel_find (chanv[i]), si->su);
      }
  }
  
  static void
  m_nick (sourceinfo_t *si, int parc, char *parv[])
  {
    server_t *s;
    bool realchange;
  
    if (parc == 8)
      {
        s = server_find (parv[5]);
        if (!s)
          {
            slog (LG_DEBUG, "m_nick(): new user on nonexistant server: %s", parv[6]);
            return;
          }
  
        slog (LG_DEBUG, "m_nick(): new user on `%s': %s", s->name, parv[0]);
  
        user_add (parv[0], parv[3], parv[4], NULL, NULL, NULL, parv[7], s, atoi (parv[2]));
  
        /* Note: cannot rely on umode +r to see if they're identified
         * -- jilles */
  
        handle_nickchange (user_find (parv[0]));
      }
  
    /* if it's only 2 then it's a nickname change */
    else if (parc == 2)
      {
        if (!si->su)
          {
            slog (LG_DEBUG, "m_nick(): server trying to change nick: %s", si->s != NULL ? si->s->name : "<none>");
            return;
          }
  
        slog (LG_DEBUG, "m_nick(): nickname change from `%s': %s", si->su->nick, parv[0]);
  
        realchange = irccasecmp (si->su->nick, parv[0]);
  
        user_changenick (si->su, parv[0], atoi (parv[1]));
  
        /* fix up +r if necessary -- jilles */
        if (realchange && !nicksvs.no_nick_ownership)
          {
            if (should_reg_umode (si->su))
              /* changed nick to registered one, reset +r */
              sts (":%s SVSMODE %s +rd %ld", nicksvs.nick, parv[0], NOW);
            else
              /* changed from registered nick, remove +r */
              sts (":%s SVSMODE %s -r+d %ld", nicksvs.nick, parv[0], NOW);
          }
  
        handle_nickchange (si->su);
      }
    else
      {
        int i;
        slog (LG_DEBUG, "m_nick(): got NICK with wrong number of params");
  
        for (i = 0; i < parc; i++)
          slog (LG_DEBUG, "m_nick():   parv[%d] = %s", i, parv[i]);
      }
  }
  
  static void
  m_quit (sourceinfo_t *si, int parc, char *parv[])
  {
    if (si->su->server == me.me)
      {
        slog (LG_DEBUG, "m_quit(): not destroying own user %s (fake direction)", si->su->nick);
        return;
      }
    slog (LG_DEBUG, "m_quit(): user leaving: %s", si->su->nick);
  
    /* user_delete() takes care of removing channels and so forth */
    user_delete (si->su);
  }
  
  static void
  m_mode (sourceinfo_t *si, int parc, char *parv[])
  {
    if (*parv[0] == '#')
      channel_mode (NULL, channel_find (parv[0]), parc - 1, &parv[1]);
    else
      user_mode (user_find (parv[0]), parv[1]);
  }
  
  static void
  m_kick (sourceinfo_t *si, int parc, char *parv[])
  {
    user_t *u = user_find (parv[1]);
    channel_t *c = channel_find (parv[0]);
  
    /* -> :rakaur KICK #shrike rintaun :test */
    slog (LG_DEBUG, "m_kick(): user was kicked: %s -> %s", parv[1], parv[0]);
  
    if (!u)
      {
        slog (LG_DEBUG, "m_kick(): got kick for nonexistant user %s", parv[1]);
        return;
      }
  
    if (!c)
      {
        slog (LG_DEBUG, "m_kick(): got kick in nonexistant channel: %s", parv[0]);
        return;
      }
  
    if (!chanuser_find (c, u))
      {
        slog (LG_DEBUG, "m_kick(): got kick for %s not in %s", u->nick, c->name);
        return;
      }
  
    chanuser_delete (c, u);
  
    /* if they kicked us, let's rejoin */
    if (is_internal_client (u))
      {
        slog (LG_DEBUG, "m_kick(): %s got kicked from %s; rejoining", u->nick, parv[0]);
        join (parv[0], u->nick);
      }
  }
  
  static void
  m_kill (sourceinfo_t *si, int parc, char *parv[])
  {
    handle_kill (si, parv[0], parc > 1 ? parv[1] : "<No reason given>");
  }
  
  static void
  m_squit (sourceinfo_t *si, int parc, char *parv[])
  {
    slog (LG_DEBUG, "m_squit(): server leaving: %s from %s", parv[0], parv[1]);
    server_delete (parv[0]);
  }
  
  static void
  m_server (sourceinfo_t *si, int parc, char *parv[])
  {
    server_t *s;
  
    slog (LG_DEBUG, "m_server(): new server: %s", parv[0]);
    s = handle_server (si, parv[0], NULL, atoi (parv[1]), parv[2]);
  
    if (s != NULL && s->uplink != me.me)
      {
        /* elicit PONG for EOB detection; pinging uplink is
         * already done elsewhere -- jilles
         */
        sts (":%s PING %s %s", me.name, me.name, s->name);
      }
  }
  
  static void
  m_stats (sourceinfo_t *si, int parc, char *parv[])
  {
    handle_stats (si->su, parv[0][0]);
  }
  
  static void
  m_admin (sourceinfo_t *si, int parc, char *parv[])
  {
    handle_admin (si->su);
  }
  
  static void
  m_version (sourceinfo_t *si, int parc, char *parv[])
  {
    handle_version (si->su);
  }
  
  static void
  m_info (sourceinfo_t *si, int parc, char *parv[])
  {
    handle_info (si->su);
  }
  
  static void
  m_whois (sourceinfo_t *si, int parc, char *parv[])
  {
    handle_whois (si->su, parv[1]);
  }
  
  static void
  m_trace (sourceinfo_t *si, int parc, char *parv[])
  {
    handle_trace (si->su, parv[0], parc >= 2 ? parv[1] : NULL);
  }
  
  static void
  m_away (sourceinfo_t *si, int parc, char *parv[])
  {
    handle_away (si->su, parc >= 1 ? parv[0] : NULL);
  }
  
  static void
  m_join (sourceinfo_t *si, int parc, char *parv[])
  {
    chanuser_t *cu;
    node_t *n, *tn;
    int chanc;
    char *chanv[256];
    int i;
  
    /* JOIN 0 is really a part from all channels */
    if (parv[0][0] == '0')
      {
        LIST_FOREACH_SAFE (n, tn, si->su->channels.head)
        {
          cu = (chanuser_t *) n->data;
          chanuser_delete (cu->chan, si->su);
        }
      }
    else
      {
        chanc = sjtoken (parv[0], ',', chanv);
        for (i = 0; i < chanc; i++)
          {
            channel_t *c = channel_find (chanv[i]);
  
            if (!c)
              {
                slog (LG_DEBUG, "m_join(): new channel: %s", parv[0]);
                c = channel_add (chanv[i], NOW, si->su->server);
                /* Tell the core to check mode locks now,
                 * otherwise it may only happen after the next
                 * mode change.
                 * DreamForge does not allow any redundant modes
                 * so this will not look ugly. -- jilles */
                /* If this is in a burst, a MODE with the
                 * simple modes will follow so we can skip
                 * this. -- jilles */
                if (!me.bursting)
                  channel_mode (NULL, c, "+");
              }
            chanuser_add (c, si->su->nick);
          }
      }
  }
  
  static void
  m_pass (sourceinfo_t *si, int parc, char *parv[])
  {
    if (strcmp (curr_uplink->pass, parv[0]))
      {
        slog (LG_INFO, "m_pass(): password mismatch from uplink; aborting");
        runflags |= RF_SHUTDOWN;
      }
  }
  
  static void
  m_error (sourceinfo_t *si, int parc, char *parv[])
  {
    slog (LG_INFO, "m_error(): error from server: %s", parv[0]);
  }
  
  static void
  m_motd (sourceinfo_t *si, int parc, char *parv[])
  {
    handle_motd (si->su);
  }
  
  static void
  nick_group (mynick_t *mn, myuser_t *mu, sourceinfo_t *si)
  {
    user_t *u;
  
    if (si->su != NULL && !irccasecmp (si->su->nick, mn->nick))
      u = si->su;
    else
      u = user_find_named (mn->nick);
  
    if (u != NULL && should_reg_umode (u))
      sts (":%s SVSMODE %s +rd %ld", nicksvs.nick, u->nick, NOW);
  }
  
  static void
  nick_ungroup (mynick_t *mn, myuser_t *mu, sourceinfo_t *si)
  {
    user_t *u;
  
    if (si->su != NULL && !irccasecmp (si->su->nick, mn->nick))
      u = si->su;
    else
      u = user_find_named (mn->nick);
  
    if (u != NULL && !nicksvs.no_nick_ownership)
      sts (":%s SVSMODE %s -r+d %ld", nicksvs.nick, u->nick, NOW);
  }
  
  static pcommand_t const pcommands[] = {
    { "PING", m_ping, 1, MSRC_USER | MSRC_SERVER },
    { "PONG", m_pong, 1, MSRC_SERVER },
    { "PRIVMSG", m_privmsg, 2, MSRC_USER },
    { "NOTICE", m_notice, 2, MSRC_UNREG | MSRC_USER | MSRC_SERVER },
    { "PART", m_part, 1, MSRC_USER },
    { "NICK", m_nick, 2, MSRC_USER | MSRC_SERVER },
    { "QUIT", m_quit, 1, MSRC_USER },
    { "MODE", m_mode, 2, MSRC_USER | MSRC_SERVER },
    { "KICK", m_kick, 2, MSRC_USER | MSRC_SERVER },
    { "KILL", m_kill, 1, MSRC_USER | MSRC_SERVER },
    { "SQUIT", m_squit, 1, MSRC_USER | MSRC_SERVER },
    { "SERVER", m_server, 3, MSRC_UNREG | MSRC_SERVER },
    { "STATS", m_stats, 2, MSRC_USER },
    { "ADMIN", m_admin, 1, MSRC_USER },
    { "VERSION", m_version, 1, MSRC_USER },
    { "INFO", m_info, 1, MSRC_USER },
    { "WHOIS", m_whois, 2, MSRC_USER },
    { "TRACE", m_trace, 1, MSRC_USER },
    { "AWAY", m_away, 0, MSRC_USER },
    { "JOIN", m_join, 1, MSRC_USER },
    { "PASS", m_pass, 1, MSRC_UNREG },
    { "ERROR", m_error, 1, MSRC_UNREG | MSRC_SERVER },
    { "TOPIC", m_topic, 4, MSRC_USER | MSRC_SERVER },
    { "MOTD", m_motd, 1, MSRC_USER },
  };
  
  dreamforge_handler::dreamforge_handler ()
  {
    mode_list = dreamforge_mode_list;
    ignore_mode_list = dreamforge_ignore_mode_list;
    status_mode_list = dreamforge_status_mode_list;
    prefix_mode_list = dreamforge_prefix_mode_list;
  
    ircd = &DreamForge;
  
    pcommand_add (&pcommands);
  
    mynick_t::callback.group.attach (nick_group);
    mynick_t::callback.ungroup.attach (nick_ungroup);
  }
  
  dreamforge_handler::~dreamforge_handler ()
  {
    mode_list = NULL;
    ignore_mode_list = NULL;
    status_mode_list = NULL;
    prefix_mode_list = NULL;
  
    ircd = NULL;
  
    mynick_t::callback.group.detach (nick_group);
    mynick_t::callback.ungroup.detach (nick_ungroup);
  
    pcommand_delete (&pcommands);
  }
} // namespace protocol
  
#define FACREG_TYPE protocol::dreamforge_handler
#define FACREG_TYPE_NAME "dreamforge"
#define FACREG_INTERFACE_TYPE protocol::handler
#include <ermyth/factory_reg.h>
