/**
 * hyperion.C: This file contains protocol support for hyperion-based ircd.
 *
 * Copyright © 2007 Pippijn van Steenhoven / The Ermyth Team
 * Rights to this code are as documented in COPYING.
 *
 *
 * Portions of this file were derived from sources bearing the following license:
 * Copyright © 2005-2007 William Pitcock, et al.
 * Rights to this code are documented in doc/pod/license.pod.
 *
 * $Id: hyperion.C,v 1.9 2007/09/22 14:27:28 pippijn Exp $
 */

/* option: use SVSLOGIN/SIGNON to remember users even if they're
 * not logged in to their current nick, etc
 */
#define USE_SVSLOGIN

#include <boost/foreach.hpp>

#include "atheme.h"
#include <util/time.h>
#include <libermyth.h>
#include "servers.h"
#include <account/mychan.h>
#include <account/mynick.h>
#include <account/myuser.h>
#include "uplink.h"
#include "pmodule.h"
#include "protocol/hyperion.h"

static char const rcsid[] = "$Id: hyperion.C,v 1.9 2007/09/22 14:27:28 pippijn Exp $";

/* *INDENT-OFF* */

namespace protocol
{
  static ircd_t Hyperion = {
    "Hyperion 1.0",                               /* IRCd name */
    "$",                                          /* TLD Prefix, used by Global. */
    false,                                        /* Whether or not we use IRCNet/TS6 UID */
    false,                                        /* Whether or not we use RCOMMAND */
    false,                                        /* Whether or not we support channel owners. */
    false,                                        /* Whether or not we support channel protection. */
    false,                                        /* Whether or not we support halfops. */
    false,                                        /* Whether or not we use P10 */
    true,                                         /* Whether or not we use vHosts. */
    CMODE_EXLIMIT | CMODE_PERM | CMODE_JUPED,     /* Oper-only cmodes */
    0,                                            /* Integer flag for owner channel flag. */
    0,                                            /* Integer flag for protect channel flag. */
    0,                                            /* Integer flag for halfops. */
    "+",                                          /* Mode we set for owner. */
    "+",                                          /* Mode we set for protect. */
    "+",                                          /* Mode we set for halfops. */
    PROTOCOL_HYPERION,                            /* Protocol type */
    CMODE_PERM | CMODE_JUPED,                     /* Permanent cmodes */
    "beIqd",                                      /* Ban-like cmodes */
    'e',                                          /* Except mchar */
    'I',                                          /* Invex mchar */
    0                                             /* Flags */
  };
  
  struct hyperion_handler : handler
  {
    hyperion_handler ();
    virtual ~hyperion_handler ();
  
    virtual unsigned int server_login (void);
    virtual void introduce_nick (user_t *u);
    virtual void invite_sts (user_t *source, user_t *target, channel_t *channel);
    virtual void quit_sts (user_t *u, char const * const reason);
    virtual void wallops_sts (char const * const text);
    virtual void join_sts (channel_t *c, user_t *u, bool isnew, char const * const modes);
    virtual void kick (char const * const from, char const * const channel, char const * const to, char const * const reason);
    virtual void privmsg (char const * const from, char const * const target, char const * const fmt, ...);
    virtual void notice_user_sts (user_t *from, user_t *target, char const * const text);
    virtual void notice_global_sts (user_t *from, char const * const mask, char const * const text);
    virtual void notice_channel_sts (user_t *from, channel_t *target, char const * const text);
    virtual void wallchops (user_t *sender, channel_t *channel, char const * const message);
    virtual void numeric_sts (char const * const from, int numeric, char const * const target, char const * const fmt, ...);
    virtual void skill (char const * const from, char const * const nick, char const * const fmt, ...);
    virtual void part_sts (channel_t *c, user_t *u);
    virtual void kline_sts (char const * const server, char const * const user, char const * const host, long duration, char const * const reason);
    virtual void unkline_sts (char const * const server, char const * const user, char const * const host);
    virtual void topic_sts (channel_t *c, char const * const setter, time_t ts, time_t prevts, char const * const topic);
    virtual void mode_sts (char const * const sender, channel_t *target, char const * const modes);
    virtual void ping_sts (void);
    virtual void ircd_on_login (char const * const origin, char const * const user, char const * const wantedhost);
    virtual bool ircd_on_logout (char const * const origin, char const * const user, char const * const wantedhost);
    virtual void jupe (char const * const server, char const * const reason);
    virtual void sethost_sts (char const * const source, char const * const target, char const * const host);
    virtual void fnc_sts (user_t *source, user_t *u, char const * const newnick, int type);
  };
  
  static cmode_t hyperion_mode_list[] = {
    { 'i', CMODE_INVITE },
    { 'm', CMODE_MOD    },
    { 'n', CMODE_NOEXT  },
    { 's', CMODE_SEC    },
    { 't', CMODE_TOPIC  },
    { 'c', CMODE_NOCOLOR},
    { 'r', CMODE_REGONLY},
    { 'R', CMODE_MODREG },
    { 'z', CMODE_OPMOD  },
    { 'g', CMODE_FINVITE},
    { 'L', CMODE_EXLIMIT},
    { 'P', CMODE_PERM   },
    { 'j', CMODE_JUPED  },
    { 'Q', CMODE_DISFWD },
    { '\0', 0 }
  };
  
  static bool check_forward(char const * const , channel_t *, mychan_t *, user_t *, myuser_t *);
  static bool check_jointhrottle(char const * const , channel_t *, mychan_t *, user_t *, myuser_t *);
  
  static extmode_t hyperion_ignore_mode_list[] = {
    /*{ 'D', 0 },*/
    { 'f', check_forward },
    { 'J', check_jointhrottle },
    { '\0', 0 }
  };
  
  static cmode_t hyperion_status_mode_list[] = {
    { 'o', CMODE_OP    },
    { 'v', CMODE_VOICE },
    { '\0', 0 }
  };
  
  static cmode_t hyperion_prefix_mode_list[] = {
    { '@', CMODE_OP    },
    { '+', CMODE_VOICE },
    { '\0', 0 }
  };
  
  static bool use_svslogin = false;
  
  /* *INDENT-ON* */
  
  static bool
  check_forward (char const * const value, channel_t *c, mychan_t *mc, user_t *u, myuser_t *mu)
  {
    channel_t *target_c;
    mychan_t *target_mc;
  
    if (*value != '#' || strlen (value) > 30)
      return false;
    if (u == NULL && mu == NULL)
      return true;
    target_c = channel_find (value);
    target_mc = mychan_t::find (value);
    if (target_c == NULL && target_mc == NULL)
      return false;
    return true;
  }
  
  static bool
  check_jointhrottle (char const * const value, channel_t *c, mychan_t *mc, user_t *u, myuser_t *mu)
  {
    char const *p = value;
    char const *arg2 = NULL;
  
    while (*p != '\0')
      {
        if (*p == ',')
          {
            if (arg2 != NULL)
              return false;
            arg2 = p + 1;
          }
        else if (!isdigit (*p))
          return false;
        p++;
      }
    if (arg2 == NULL)
      return false;
    if (p - arg2 > 5 || arg2 - value - 1 > 5 || !atoi (value) || !atoi (arg2))
      return false;
    return true;
  }
  
  node_t *
  hyperion_next_matching_ban (channel_t *c, user_t *u, int type, node_t *first)
  {
    chanban_t *cb;
    node_t *n;
    char hostbuf[NICKLEN + USERLEN + HOSTLEN];
    char realbuf[NICKLEN + USERLEN + HOSTLEN];
    char ipbuf[NICKLEN + USERLEN + HOSTLEN];
  
    snprintf (hostbuf, sizeof hostbuf, "%s!%s@%s", u->nick, u->user, u->vhost);
    snprintf (realbuf, sizeof realbuf, "%s!%s@%s", u->nick, u->user, u->host);
    /* will be nick!user@ if ip unknown, doesn't matter */
    snprintf (ipbuf, sizeof ipbuf, "%s!%s@%s", u->nick, u->user, u->ip);
    LIST_FOREACH (n, first)
    {
      cb = static_cast<chanban_t *> (n->data);
  
      if (cb->type == type && (!match (cb->mask, hostbuf) || !match (cb->mask, realbuf) || !match (cb->mask, ipbuf)))
        return n;
      if (cb->type == 'd' && type == 'b' && !match (cb->mask, u->gecos))
        return n;
    }
    return NULL;
  }
  
  /* login to our uplink */
  unsigned int
  hyperion_handler::server_login (void)
  {
    int ret;
  
    ret = sts ("PASS %s :TS", curr_uplink->pass);
    if (ret == 1)
      return 1;
  
    me.bursting = true;
  
    sts ("CAPAB :QS EX DE CHW IE QU DNCR SRV SIGNON");
    sts ("SERVER %s 1 :%s", me.name, me.desc);
    sts ("SVINFO 5 3 0 :%ld", NOW);
  
    services_init ();
  
    return 0;
  }
  
  /* introduce a nick.
   * Hyperion sucks because it has to do things differently.
   * Thus we have to make all clients opered, because there is no real way to get
   * around the problem.
   *
   * Protocol information ripped from theia/dancer-services (Hybserv2 TS of some sort).
   *    -- nenolod
   */
  void
  hyperion_handler::introduce_nick (user_t *u)
  {
    char const * const privs = "6@BFmMopPRUX";
    sts ("NICK %s 1 %ld +ei%s %s %s %s 0.0.0.0 :%s", u->nick, u->ts, privs, u->user, u->host, me.name, u->gecos);
    sts (":%s OPER %s +%s", me.name, u->nick, privs);
  }
  
  /* invite a user to a channel */
  void
  hyperion_handler::invite_sts (user_t *sender, user_t *target, channel_t *channel)
  {
    sts (":%s INVITE %s %s", sender->nick, target->nick, channel->name);
  }
  
  void
  hyperion_handler::quit_sts (user_t *u, char const * const reason)
  {
    if (!me.connected)
      return;
  
    sts (":%s QUIT :%s", u->nick, reason);
  }
  
  /* WALLOPS wrapper */
  void
  hyperion_handler::wallops_sts (char const * const text)
  {
    /* Generate +s server notice -- jilles */
    sts (":%s WALLOPS 1-0 :%s", me.name, text);
  }
  
  /* join a channel */
  void
  hyperion_handler::join_sts (channel_t *c, user_t *u, bool isnew, char const * const modes)
  {
    if (isnew)
      sts (":%s SJOIN %ld %s %s :@%s", me.name, c->ts, c->name, modes, u->nick);
    else
      sts (":%s SJOIN %ld %s + :@%s", me.name, c->ts, c->name, u->nick);
  }
  
  /* kicks a user from a channel */
  void
  hyperion_handler::kick (char const * const from, char const * const channel, char const * const to, char const * const reason)
  {
    channel_t *chan = channel_find (channel);
    user_t *user = user_find (to);
  
    if (!chan || !user)
      return;
  
    sts (":%s KICK %s %s :%s", from, channel, to, reason);
  
    chanuser_delete (chan, user);
  }
  
  /* PRIVMSG wrapper */
  void
  hyperion_handler::privmsg (char const * const from, char const * const target, char const * const fmt, ...)
  {
    va_list ap;
    char buf[BUFSIZE];
  
    va_start (ap, fmt);
    vsnprintf (buf, BUFSIZE, fmt, ap);
    va_end (ap);
  
    sts (":%s PRIVMSG %s :%s", from, target, buf);
  }
  
  /* NOTICE wrapper */
  void
  hyperion_handler::notice_user_sts (user_t *from, user_t *target, char const * const text)
  {
    sts (":%s NOTICE %s :%s", from ? from->nick : me.name, target->nick, text);
  }
  
  void
  hyperion_handler::notice_global_sts (user_t *from, char const * const mask, char const * const text)
  {
    sts (":%s NOTICE %s%s :%s", from ? from->nick : me.name, ircd->tldprefix, mask, text);
  }
  
  void
  hyperion_handler::notice_channel_sts (user_t *from, channel_t *target, char const * const text)
  {
    sts (":%s NOTICE %s :%s", from ? from->nick : me.name, target->name, text);
  }
  
  void
  hyperion_handler::wallchops (user_t *sender, channel_t *channel, char const * const message)
  {
    /* +p does not grant the ability to send to @#channel (!) */
    if (chanuser_find (channel, sender))
      sts (":%s NOTICE @%s :%s", CLIENT_NAME (sender), channel->name, message);
    else                                /* do not join for this, everyone would see -- jilles */
      handler::wallchops (sender, channel, message);
  }
  
  /* numeric wrapper */
  void
  hyperion_handler::numeric_sts (char const * const from, int numeric, char const * const target, char const * const fmt, ...)
  {
    va_list ap;
    char buf[BUFSIZE];
  
    va_start (ap, fmt);
    vsnprintf (buf, BUFSIZE, fmt, ap);
    va_end (ap);
  
    sts (":%s %d %s %s", from, numeric, target, buf);
  }
  
  /* KILL wrapper */
  void
  hyperion_handler::skill (char const * const from, char const * const nick, char const * const fmt, ...)
  {
    va_list ap;
    char buf[BUFSIZE];
  
    va_start (ap, fmt);
    vsnprintf (buf, BUFSIZE, fmt, ap);
    va_end (ap);
  
  #if 0
    sts (":%s KILL %s :%s!%s!%s (%s)", from, nick, from, from, from, buf);
  #else
    /* Use COLLIDE to cut down on server notices.
     * The current version of hyperion does not do COLLIDE reasons,
     * so fake it.
     */
    sts (":%s NOTICE %s :*** Disconnecting you (%s (%s))", me.name, nick, from, buf);
    sts (":%s COLLIDE %s :(%s)", me.name, nick, buf);
  #endif
  }
  
  /* PART wrapper */
  void
  hyperion_handler::part_sts (channel_t *c, user_t *u)
  {
    sts (":%s PART %s", u->nick, c->name);
  }
  
  /* server-to-server KLINE wrapper */
  void
  hyperion_handler::kline_sts (char const * const server, char const * const user, char const * const host, long duration, char const * const reason)
  {
    if (!me.connected)
      return;
  
    if (duration)
      sts (":%s KLINE %s %ld %s@%s :%s", me.name, me.name, duration > 60 ? (duration / 60) : 1, user, host, reason);
    else
      sts (":%s KLINE %s %s@%s :%s", me.name, me.name, user, host, reason);
  }
  
  /* server-to-server UNKLINE wrapper */
  void
  hyperion_handler::unkline_sts (char const * const server, char const * const user, char const * const host)
  {
    if (!me.connected)
      return;
  
    sts (":%s UNKLINE %s@%s %ld", opersvs.nick, user, host, NOW);
  }
  
  /* topic wrapper */
  void
  hyperion_handler::topic_sts (channel_t *c, char const * const setter, time_t ts, time_t prevts, char const * const topic)
  {
    if (!me.connected || !c)
      return;
  
    /* Send 0 channelts so this will always be accepted */
    sts (":%s STOPIC %s %s %ld 0 :%s", chansvs.nick, c->name, setter, ts, topic);
  }
  
  /* mode wrapper */
  void
  hyperion_handler::mode_sts (char const * const sender, channel_t *target, char const * const modes)
  {
    if (!me.connected)
      return;
  
    sts (":%s MODE %s %s", sender, target->name, modes);
  }
  
  /* ping wrapper */
  void
  hyperion_handler::ping_sts (void)
  {
    if (!me.connected)
      return;
  
    sts ("PING :%s", me.name);
  }
  
  /* protocol-specific stuff to do on login */
  void
  hyperion_handler::ircd_on_login (char const * const origin, char const * const user, char const * const wantedhost)
  {
    user_t *u;
  
    if (!me.connected)
      return;
  
    u = user_find (origin);
    if (!u)
      return;
    if (use_svslogin)
      sts (":%s SVSLOGIN %s %s %s %s %s %s", me.name, u->server->name, origin, user, origin, u->user, wantedhost ? wantedhost : u->vhost);
    /* we'll get a SIGNON confirming the changes later, no need
     * to change the fields yet */
  
    /* set +e if they're identified to the nick they are using */
    if (should_reg_umode (u))
      sts (":%s MODE %s +e", me.name, origin);
  }
  
  /* protocol-specific stuff to do on login */
  bool
  hyperion_handler::ircd_on_logout (char const * const origin, char const * const user, char const * const wantedhost)
  {
    user_t *u;
  
    if (!me.connected)
      return false;
  
    u = user_find (origin);
    if (!u)
      return false;
    if (use_svslogin)
      sts (":%s SVSLOGIN %s %s %s %s %s %s", me.name, u->server->name, origin, "0", origin, u->user, wantedhost ? u->host : u->vhost);
  
    if (!nicksvs.no_nick_ownership)
      sts (":%s MODE %s -e", me.name, origin);
  
    return false;
  }
  
  void
  hyperion_handler::jupe (char const * const server, char const * const reason)
  {
    if (!me.connected)
      return;
  
    server_delete (server);
    sts (":%s SQUIT %s :%s", opersvs.nick, server, reason);
    sts (":%s SERVER %s 2 :%s", me.name, server, reason);
  }
  
  void
  hyperion_handler::sethost_sts (char const * const source, char const * const target, char const * const host)
  {
    if (!me.connected)
      return;
  
    sts (":%s SETHOST %s :%s", source, target, host);
  }
  
  void
  hyperion_handler::fnc_sts (user_t *source, user_t *u, char const * const newnick, int type)
  {
    /* XXX this should be combined with the SVSLOGIN to set login id
     * and SETHOST, if any -- jilles */
    sts (":%s SVSLOGIN %s %s %s %s %s %s", me.name, u->server->name, u->nick, u->myuser ? u->myuser->name : "0", newnick, u->user, u->vhost);
  }
  
  static void
  m_topic (sourceinfo_t *si, int parc, char *parv[])
  {
    channel_t *c = channel_find (parv[0]);
  
    if (!c || !si->su)
      return;
  
    handle_topic_from (si, c, si->su->nick, NOW, parv[1]);
  }
  
  static void
  m_stopic (sourceinfo_t *si, int parc, char *parv[])
  {
    channel_t *c = channel_find (parv[0]);
    time_t channelts;
    time_t topicts;
  
    if (c == NULL)
      return;
  
    /* Our uplink is trying to change the topic during burst,
     * and we have already set a topic. Assume our change won.
     * -- jilles */
    if (si->s != NULL && si->s->uplink == me.me && !(si->s->flags & SF_EOB) && c->topic != NULL)
      return;
  
    /* hyperion will propagate an STOPIC even if it's not applied
     * locally :( */
    topicts = atol (parv[2]);
    channelts = atol (parv[3]);
    if (c->topic == NULL || ((strcmp (c->topic, parv[1]) && channelts < c->ts) || (channelts == c->ts && topicts > c->topicts)))
      handle_topic_from (si, c, parv[1], topicts, parv[parc - 1]);
  }
  
  static void
  m_ping (sourceinfo_t *si, int parc, char *parv[])
  {
    /* reply to PING's */
    sts (":%s PONG %s %s", me.name, me.name, parv[0]);
  }
  
  static void
  m_pong (sourceinfo_t *si, int parc, char *parv[])
  {
    server_t *s;
  
    /* someone replied to our PING */
    if (!parv[0])
      return;
    s = server_find (parv[0]);
    if (s == NULL)
      return;
    handle_eob (s);
  
    if (irccasecmp (me.actual, parv[0]))
      return;
  
    me.uplinkpong = NOW;
  
    /* -> :test.projectxero.net PONG test.projectxero.net :shrike.malkier.net */
    if (me.bursting)
      {
  #ifdef HAVE_GETTIMEOFDAY
        e_time (burstime, &burstime);
  
        slog (LG_INFO, "m_pong(): finished synching with uplink (%d %s)", (tv2ms (&burstime) > 1000) ? (tv2ms (&burstime) / 1000) : tv2ms (&burstime), (tv2ms (&burstime) > 1000) ? "s" : "ms");
  
        wallops ("Finished synching to network in %d %s.", (tv2ms (&burstime) > 1000) ? (tv2ms (&burstime) / 1000) : tv2ms (&burstime), (tv2ms (&burstime) > 1000) ? "s" : "ms");
  #else
        slog (LG_INFO, "m_pong(): finished synching with uplink");
        wallops ("Finished synching to network.");
  #endif
  
        me.bursting = false;
      }
  }
  
  static void
  m_privmsg (sourceinfo_t *si, int parc, char *parv[])
  {
    if (parc != 2)
      return;
  
    handle_message (si, parv[0], false, parv[1]);
  }
  
  static void
  m_notice (sourceinfo_t *si, int parc, char *parv[])
  {
    if (parc != 2)
      return;
  
    handle_message (si, parv[0], true, parv[1]);
  }
  
  static void
  m_sjoin (sourceinfo_t *si, int parc, char *parv[])
  {
    /* -> :proteus.malkier.net SJOIN 1073516550 #shrike +tn :@sycobuny @+rakaur */
  
    channel_t *c;
    unsigned int userc;
    char *userv[256];
    unsigned int i;
    time_t ts;
  
    /* :origin SJOIN ts chan modestr [key or limits] :users */
    c = channel_find (parv[1]);
    ts = atol (parv[0]);
  
    if (!c)
      {
        slog (LG_DEBUG, "m_sjoin(): new channel: %s", parv[1]);
        c = channel_add (parv[1], ts, si->s);
      }
  
    if (ts < c->ts)
      {
        chanuser_t *cu;
        node_t *n;
  
        /* the TS changed.  a TS change requires the following things
         * to be done to the channel:  reset all modes to nothing, remove
         * all status modes on known users on the channel (including ours),
         * and set the new TS.
         */
  
        clear_simple_modes (c);
  
        LIST_FOREACH (n, c->members.head)
        {
          cu = (chanuser_t *) n->data;
          if (cu->user->server == me.me)
            {
              /* it's a service, reop */
              sts (":%s MODE %s +o %s", cu->user->nick, c->name, cu->user->nick);
              cu->modes = CMODE_OP;
            }
          else
            cu->modes = 0;
        }
  
        slog (LG_DEBUG, "m_sjoin(): TS changed for %s (%ld -> %ld)", c->name, c->ts, ts);
  
        c->ts = ts;
        c->callback.tschange (c);
      }
  
    channel_mode (NULL, c, parc - 3, parv + 2);
  
    userc = sjtoken (parv[parc - 1], ' ', userv);
  
    for (i = 0; i < userc; i++)
      chanuser_add (c, userv[i]);
  
    if (c->nummembers == 0 && !(c->modes & ircd->perm_mode))
      channel_delete (c);
  }
  
  static void
  m_part (sourceinfo_t *si, int parc, char *parv[])
  {
    int chanc;
    char *chanv[256];
    int i;
  
    chanc = sjtoken (parv[0], ',', chanv);
    for (i = 0; i < chanc; i++)
      {
        slog (LG_DEBUG, "m_part(): user left channel: %s -> %s", si->su->nick, chanv[i]);
  
        chanuser_delete (channel_find (chanv[i]), si->su);
      }
  }
  
  static void
  m_nick (sourceinfo_t *si, int parc, char *parv[])
  {
    server_t *s;
    user_t *u;
    bool realchange;
  
    /* got the right number of args for an introduction? */
    if (parc == 9)
      {
        s = server_find (parv[6]);
        if (!s)
          {
            slog (LG_DEBUG, "m_nick(): new user on nonexistant server: %s", parv[6]);
            return;
          }
  
        slog (LG_DEBUG, "m_nick(): new user on `%s': %s", s->name, parv[0]);
  
        u = user_add (parv[0], parv[4], parv[5], NULL, parv[7], NULL, parv[8], s, atoi (parv[2]));
  
        user_mode (u, parv[3]);
  
        /* umode +e: identified to current nick */
        /* As hyperion clears +e on nick changes, this is safe. */
        if (!use_svslogin && strchr (parv[3], 'e'))
          handle_burstlogin (u, NULL);
  
        /* if the server supports SIGNON, we will get an SNICK
         * for this user, potentially with a login name
         */
        if (!use_svslogin)
          handle_nickchange (u);
      }
  
    /* if it's only 2 then it's a nickname change */
    else if (parc == 2)
      {
        if (!si->su)
          {
            slog (LG_DEBUG, "m_nick(): server trying to change nick: %s", si->s != NULL ? si->s->name : "<none>");
            return;
          }
  
        slog (LG_DEBUG, "m_nick(): nickname change from `%s': %s", si->su->nick, parv[0]);
  
        realchange = irccasecmp (si->su->nick, parv[0]);
  
        user_changenick (si->su, parv[0], atoi (parv[1]));
  
        /* fix up +e if necessary -- jilles */
        if (realchange && should_reg_umode (si->su))
          /* changed nick to registered one, reset +e */
          sts (":%s MODE %s +e", me.name, parv[0]);
  
        handle_nickchange (si->su);
      }
    else
      {
        int i;
        slog (LG_DEBUG, "m_nick(): got NICK with wrong number of params");
  
        for (i = 0; i < parc; i++)
          slog (LG_DEBUG, "m_nick():   parv[%d] = %s", i, parv[i]);
      }
  }
  
  static void
  m_quit (sourceinfo_t *si, int parc, char *parv[])
  {
    slog (LG_DEBUG, "m_quit(): user leaving: %s", si->su->nick);
  
    /* user_delete() takes care of removing channels and so forth */
    user_delete (si->su);
  }
  
  static void
  m_mode (sourceinfo_t *si, int parc, char *parv[])
  {
    if (*parv[0] == '#')
      channel_mode (NULL, channel_find (parv[0]), parc - 1, &parv[1]);
    else
      user_mode (user_find (parv[0]), parv[1]);
  }
  
  static void
  m_kick (sourceinfo_t *si, int parc, char *parv[])
  {
    user_t *u = user_find (parv[1]);
    channel_t *c = channel_find (parv[0]);
  
    /* -> :rakaur KICK #shrike rintaun :test */
    slog (LG_DEBUG, "m_kick(): user was kicked: %s -> %s", parv[1], parv[0]);
  
    if (!u)
      {
        slog (LG_DEBUG, "m_kick(): got kick for nonexistant user %s", parv[1]);
        return;
      }
  
    if (!c)
      {
        slog (LG_DEBUG, "m_kick(): got kick in nonexistant channel: %s", parv[0]);
        return;
      }
  
    if (!chanuser_find (c, u))
      {
        slog (LG_DEBUG, "m_kick(): got kick for %s not in %s", u->nick, c->name);
        return;
      }
  
    chanuser_delete (c, u);
  
    /* if they kicked us, let's rejoin */
    if (is_internal_client (u))
      {
        slog (LG_DEBUG, "m_kick(): %s got kicked from %s; rejoining", u->nick, parv[0]);
        join (parv[0], u->nick);
      }
  }
  
  static void
  m_kill (sourceinfo_t *si, int parc, char *parv[])
  {
    /* serves for both KILL and COLLIDE
     * COLLIDE only originates from servers and may not have
     * a reason field, but the net effect is identical
     * -- jilles */
    handle_kill (si, parv[0], parc > 1 ? parv[1] : "<No reason given>");
  }
  
  static void
  m_squit (sourceinfo_t *si, int parc, char *parv[])
  {
    slog (LG_DEBUG, "m_squit(): server leaving: %s from %s", parv[0], parv[1]);
    server_delete (parv[0]);
  }
  
  static void
  m_server (sourceinfo_t *si, int parc, char *parv[])
  {
    server_t *s;
  
    slog (LG_DEBUG, "m_server(): new server: %s", parv[0]);
    s = handle_server (si, parv[0], NULL, atoi (parv[1]), parv[2]);
  
    if (s != NULL && s->uplink != me.me)
      {
        /* elicit PONG for EOB detection; pinging uplink is
         * already done elsewhere -- jilles
         */
        sts (":%s PING %s %s", me.name, me.name, s->name);
      }
  }
  
  static void
  m_stats (sourceinfo_t *si, int parc, char *parv[])
  {
    handle_stats (si->su, parv[0][0]);
  }
  
  static void
  m_admin (sourceinfo_t *si, int parc, char *parv[])
  {
    handle_admin (si->su);
  }
  
  static void
  m_version (sourceinfo_t *si, int parc, char *parv[])
  {
    handle_version (si->su);
  }
  
  static void
  m_info (sourceinfo_t *si, int parc, char *parv[])
  {
    handle_info (si->su);
  }
  
  static void
  m_whois (sourceinfo_t *si, int parc, char *parv[])
  {
    handle_whois (si->su, parv[1]);
  }
  
  static void
  m_trace (sourceinfo_t *si, int parc, char *parv[])
  {
    handle_trace (si->su, parv[0], parc >= 2 ? parv[1] : NULL);
  }
  
  static void
  m_away (sourceinfo_t *si, int parc, char *parv[])
  {
    handle_away (si->su, parc >= 1 ? parv[0] : NULL);
  }
  
  static void
  m_join (sourceinfo_t *si, int parc, char *parv[])
  {
    chanuser_t *cu;
    node_t *n, *tn;
  
    /* JOIN 0 is really a part from all channels */
    if (parv[0][0] == '0')
      {
        LIST_FOREACH_SAFE (n, tn, si->su->channels.head)
        {
          cu = (chanuser_t *) n->data;
          chanuser_delete (cu->chan, si->su);
        }
      }
  }
  
  static void
  m_pass (sourceinfo_t *si, int parc, char *parv[])
  {
    if (strcmp (curr_uplink->pass, parv[0]))
      {
        slog (LG_INFO, "m_pass(): password mismatch from uplink; aborting");
        runflags |= RF_SHUTDOWN;
      }
  }
  
  static void
  m_error (sourceinfo_t *si, int parc, char *parv[])
  {
    slog (LG_INFO, "m_error(): error from server: %s", parv[0]);
  }
  
  static void
  m_capab (sourceinfo_t *si, int parc, char *parv[])
  {
    char *p;
  
  #ifdef USE_SVSLOGIN
    for (p = strtok (parv[0], " "); p != NULL; p = strtok (NULL, " "))
      {
        if (!irccasecmp (p, "SIGNON"))
          {
            slog (LG_DEBUG, "m_capab(): uplink can do SIGNON, enabling support.");
            use_svslogin = true;
          }
      }
  #endif
  }
  
  static void
  m_snick (sourceinfo_t *si, int parc, char *parv[])
  {
    user_t *u;
  
    /* SNICK <nick> <orignick> <spoofhost> <firsttime> <dnshost> <servlogin> */
    u = user_find (parv[0]);
  
    if (!u)
      return;
  
    if (strcmp (u->host, parv[2]))        /* User is not using spoofhost, assume no I:line spoof */
      {
        strlcpy (u->vhost, parv[4], HOSTLEN);
      }
  
    if (use_svslogin)
      {
        if (parc >= 6)
          if (strcmp (parv[5], "0"))
            handle_burstlogin (u, parv[5]);
  
        handle_nickchange (u);
      }
  }
  
  static void
  m_sethost (sourceinfo_t *si, int parc, char *parv[])
  {
    user_t *u;
  
    /* SETHOST <nick> <newhost> */
    u = user_find (parv[0]);
  
    if (!u)
      return;
  
    strlcpy (u->vhost, parv[1], HOSTLEN);
  }
  
  static void
  m_setident (sourceinfo_t *si, int parc, char *parv[])
  {
    user_t *u;
  
    /* SETHOST <nick> <newident> */
    u = user_find (parv[0]);
  
    if (!u)
      return;
  
    strlcpy (u->user, parv[1], USERLEN);
  }
  
  static void
  m_setname (sourceinfo_t *si, int parc, char *parv[])
  {
    user_t *u;
  
    /* SETNAME <nick> <newreal> */
    u = user_find (parv[0]);
  
    if (!u)
      return;
  
    strlcpy (u->gecos, parv[1], GECOSLEN);
  }
  
  /*
   * Make broadcasted changes to a user's login id, nick, user, and hostname.
   *
   * parv[0] = login id
   * parv[1] = new nick
   * parv[2] = new ident
   * parv[3] = new hostname
   * parv[4] = ts
   */
  static void
  m_signon (sourceinfo_t *si, int parc, char *parv[])
  {
    char *nick_parv[2];
  
    slog (LG_DEBUG, "m_signon(): signon %s -> %s!%s@%s (login %s)", si->su->nick, parv[1], parv[2], parv[3], parv[0]);
  
    strlcpy (si->su->user, parv[2], USERLEN);
    strlcpy (si->su->vhost, parv[3], HOSTLEN);
    nick_parv[0] = parv[1];
    nick_parv[1] = parv[4];
    if (strcmp (si->su->nick, parv[1]))
      m_nick (si, 2, nick_parv);
    /* don't use login id, assume everyone signs in via the services */
  }
  
  static void
  m_motd (sourceinfo_t *si, int parc, char *parv[])
  {
    handle_motd (si->su);
  }
  
  static void
  nick_group (mynick_t *mn, myuser_t *mu, sourceinfo_t *si)
  {
    user_t *u;
  
    if (si->su != NULL && !irccasecmp (si->su->nick, mn->nick))
      u = si->su;
    else
      u = user_find_named (mn->nick);
  
    if (u != NULL && should_reg_umode (u))
      sts (":%s MODE %s +e", me.name, u->nick);
  }
  
  static void
  nick_ungroup (mynick_t *mn, myuser_t *mu, sourceinfo_t *si)
  {
    user_t *u;
  
    if (si->su != NULL && !irccasecmp (si->su->nick, mn->nick))
      u = si->su;
    else
      u = user_find_named (mn->nick);
  
    if (u != NULL && !nicksvs.no_nick_ownership)
      sts (":%s MODE %s -e", me.name, u->nick);
  }
  
  static pcommand_t const pcommands[] = {
    { "PING", m_ping, 1, MSRC_USER | MSRC_SERVER },
    { "PONG", m_pong, 1, MSRC_SERVER },
    { "PRIVMSG", m_privmsg, 2, MSRC_USER },
    { "NOTICE", m_notice, 2, MSRC_UNREG | MSRC_USER | MSRC_SERVER },
    { "SJOIN", m_sjoin, 4, MSRC_SERVER },
    { "PART", m_part, 1, MSRC_USER },
    { "NICK", m_nick, 2, MSRC_USER | MSRC_SERVER },
    { "QUIT", m_quit, 1, MSRC_USER },
    { "MODE", m_mode, 2, MSRC_USER | MSRC_SERVER },
    { "KICK", m_kick, 2, MSRC_USER | MSRC_SERVER },
    { "REMOVE", m_kick, 2, MSRC_USER | MSRC_SERVER },        /* same net result */
    { "KILL", m_kill, 1, MSRC_USER | MSRC_SERVER },
    { "COLLIDE", m_kill, 1, MSRC_SERVER },        /* same net result */
    { "SQUIT", m_squit, 1, MSRC_USER | MSRC_SERVER },
    { "SERVER", m_server, 3, MSRC_UNREG | MSRC_SERVER },
    { "STATS", m_stats, 2, MSRC_USER },
    { "ADMIN", m_admin, 1, MSRC_USER },
    { "VERSION", m_version, 1, MSRC_USER },
    { "INFO", m_info, 1, MSRC_USER },
    { "WHOIS", m_whois, 2, MSRC_USER },
    { "TRACE", m_trace, 1, MSRC_USER },
    { "AWAY", m_away, 0, MSRC_USER },
    { "JOIN", m_join, 1, MSRC_USER },
    { "PASS", m_pass, 1, MSRC_UNREG },
    { "ERROR", m_error, 1, MSRC_UNREG | MSRC_SERVER },
    { "TOPIC", m_topic, 2, MSRC_USER },
    { "STOPIC", m_stopic, 5, MSRC_USER | MSRC_SERVER },
    { "SNICK", m_snick, 5, MSRC_SERVER },
    { "SETHOST", m_sethost, 2, MSRC_USER | MSRC_SERVER },
    { "SETIDENT", m_setident, 2, MSRC_USER | MSRC_SERVER },
    { "SETNAME", m_setname, 2, MSRC_USER | MSRC_SERVER },
    { "SIGNON", m_signon, 5, MSRC_USER },
    { "CAPAB", m_capab, 1, MSRC_UNREG },
    { "MOTD", m_motd, 1, MSRC_USER },
  };
  
  hyperion_handler::hyperion_handler ()
  {
    mode_list = hyperion_mode_list;
    ignore_mode_list = hyperion_ignore_mode_list;
    status_mode_list = hyperion_status_mode_list;
    prefix_mode_list = hyperion_prefix_mode_list;
  
    ircd = &Hyperion;
  
    pcommand_add (&pcommands);
  
    mynick_t::callback.group.attach (nick_group);
    mynick_t::callback.ungroup.attach (nick_ungroup);
  }
  
  hyperion_handler::~hyperion_handler ()
  {
    mode_list = NULL;
    ignore_mode_list = NULL;
    status_mode_list = NULL;
    prefix_mode_list = NULL;
  
    ircd = NULL;
  
    mynick_t::callback.group.detach (nick_group);
    mynick_t::callback.ungroup.detach (nick_ungroup);
  
    pcommand_delete (&pcommands);
  }
} // namespace protocol
  
#define FACREG_TYPE protocol::hyperion_handler
#define FACREG_TYPE_NAME "hyperion"
#define FACREG_INTERFACE_TYPE protocol::handler
#include <ermyth/factory_reg.h>
