package Ermyth::Accessor;

use strict;
use warnings FATAL => 'all';
use utf8;

my $rcsid = '$Id: Accessor.pm,v 1.4 2007/07/28 17:26:27 pippijn Exp $';

use base "Exporter";

our @EXPORT = qw(acc_rw acc_ro acc_wo regservice);

=head1 Ermyth::Accessor

Create accessors for object variables.

=head1 SYNOPSIS

   package MyClass;

   use Ermyth::Accessor;

   acc_ro 'lastarg';

   sub new {
      my ($class) = @_;

      bless {
         lastarg => pop,
      }, $class;
   }

=cut

sub _acc_vars {
   my ($glob, $package) = (shift, caller 1);
   defined $package or die "no caller";
   die "bareword required (arg 1)"
      if ref $glob || $glob eq '';
   my ($var) = @_ ? shift : $glob;
   die "bareword required (arg 2)"
      if ref $var || $var eq '';

   $package, $glob, $var
}

sub _acc_make {
   my ($name, $sub) = @_;
   no strict 'refs';

   *$name = $sub
}

=head1 METHODS

=over

=item acc_rw (*, *)

   acc_rw 'variable', 'name';

Create read-write accessors that can be used as:

   $ob->name ("newval"); # setting a new value to variable
   my $val = $ob->name; # getting the value of variable

=cut

sub acc_rw (*;*) {
   my ($package, $glob, $var) = &_acc_vars;
   _acc_make "${package}::$glob", sub {
      my $self = shift;
      $self->{$var} = shift
         if @_;
      $self->{$var}
   }
}

=item acc_ro (*, *)

   acc_ro 'variable', 'name';

Create read-only accessors that can be used as:

   my $val = $ob->name; # getting the value of variable

=cut

sub acc_ro (*;*) {
   my ($package, $glob, $var) = &_acc_vars;
   _acc_make "${package}::$glob", sub {
      shift()->{$var}
   }
}

=item acc_wo (*, *)

   acc_wo 'variable', 'name';

Create write-only accessors that can be used as:

   $ob->name ("newval"); # setting a new value to variable

=cut

sub acc_wo (*;*) {
   my ($package, $glob, $var) = &_acc_vars;
   _acc_make "${package}::$glob", sub {
      my $self = shift;
      die "usage: <value>"
         unless @_;
      $self->{$var} = shift;
   }
}

sub _create_service ($) {
   my $package = $_[0];
   # This is the new() method for each service
   _acc_make "${package}::new", sub {
      my ($class, $ermyth, $svsname) = @_;
      my $self = bless {
         base => $ermyth,
         name => $svsname,
      }, $package;

      _acc_make "${package}::base", sub { shift()->{base} };
      _acc_make "${package}::name", sub { shift()->{name} };
      
      my @cmds = split /\n/, $self->base->command ($self, "HELP", "COMMANDS");
      shift @cmds; shift @cmds; pop @cmds; # Get rid of the first two lines and the last one
      while (my $cmd = pop @cmds) {
         $cmd =~ s/^(\w+).+/$1/;
         $cmd = lc $cmd;
         # Make accessors for commands at runtime
         _acc_make "${package}::$cmd", sub {
            my $self = shift;
            $self->base->command ($self, uc $cmd, @_)
         }
      }
      
      $self
   }  
}

=item regservice (*someserv, *SomeServ)

Used internally by Ermyth to reduce code duplication.
This registers a service C<service> that can be accessed as
C<$ermyth-E<gt>someserv>.

=back

=cut

sub regservice (*;*) {
   my ($package, $glob) = &_acc_vars;
   my $svs = "${package}::$glob";
   my $var = lc $glob;
   _create_service $svs;
   _acc_make "${package}::$var", sub {
      $_[0]->{$glob} = new $svs ($_[0], $glob)
         unless $_[0]->{$glob};
      $_[0]->{$glob}
   }
}

=head1 AUTHOR

Copyright E<xa9> 2007 Pippijn van Steenhoven

=head1 LICENSE

This library is free software, you can redistribute it and/or modify
it under the same terms as Perl itself.

=cut

1
