package Ermyth::Web::Model::Pod;

use strict;
use warnings FATAL => 'all';
use utf8;

my $rcsid = '$Id: Pod.pm,v 1.5 2007/07/28 17:26:27 pippijn Exp $';

use base qw/Catalyst::Model/;
use NEXT;
use Carp;

use IO::Dir;
use Path::Class (); # () to prevent file to be imported
use IO::File;
use Pod::POM;

our $VERSION = 0.01;

=head1 NAME

Ermyth::Web::Model::Pod

=head1 SYNOPSIS

In your application configuration (where you 'use Catalyst'):

   __PACKAGE__->config (
      'Model::Pod' => {
         root_dir => __PACKAGE__->path_to ('root/pod'),
      },
   );

In your Controller:

   sub help_on_login : Local {
      my ($self, $c) = @_;

      $c->stash->{pod}      = $c->model('Pod')->html ("help.pod", 'login');
      $c->stash->{template} = 'pod.tt';
   }

This is what help.pod could look like:

   =head1 login

   =head2 Help on Login

   To log in, you need to provide your username and password.

   =head1 standard

   =head2 Help

   Here you can get help.

And this might be pod.tt:

   [% pod %]

In this Model, =head1 means 'section' and can be chosen with the second
argument, which defaults to 'standard'.

=head1 DESCRIPTION

This is a POD formatting model.

=head1 METHODS

=over

=item new

Initialises the formatter.

=cut

sub new {
   my $self = shift->NEXT::new(@_);
   
   croak "->config->{root_dir} must be defined for this model\n"
      unless $self->{root_dir};
   
   unless (ref $self->{root_dir} ) {
      # If a string is provided turn into a Path::Class
      $self->{root_dir} = Path::Class::dir($self->{root_dir})
   }
   
   $self->{dir_create_mask} ||= 0775;
   $self->{root_dir}->mkpath(0, $self->{dir_mask});
   $self->{directory} = Path::Class::dir('/');
   $self->{_dir} = $self->{root_dir};
   $self->{parser} = new Pod::POM;

   
   $self
}

=item $self->parser

Returns the Pod::POM instance.

=cut

sub parser { $_[0]->{parser} }

=item $self->file ($file)

Used internally to make sure the file exists.

=cut

sub file {
   my ($self, $file) = @_;
   
   return
      unless $file;
   
   $file = (ref $file ? $file : Path::Class::file ($file))->absolute ($self->{_dir});
   
   return undef
      unless $self->{root_dir}->subsumes($file);
   
   # Make sure the dir tree exists
   $file->dir->mkpath(0, $self->{dir_create_mask});
   return $file;
   
}

=item $self->html ($file, $caption)

Format the content of head1 named $caption in $file to HTML.

=cut

sub html {
   my ($self, $file, $caption) = @_;
   $file = $self->file ($file)
      or return undef;
   $file->stat
      or return undef;
   $caption = "standard"
      unless $caption;

   my $pod = $file->slurp (@_);
   my ($title) = grep { $_->title eq $caption } $self->parser->parse_text($pod)->head1;

   return $title->content->present ("Ermyth::Web::Model::Pod::HTML");
}


package Ermyth::Web::Model::Pod::HTML;

use strict;
use warnings;
use utf8;

sub view_pod {
   my ($self, $item) = @_;
   $item->content->present ($self)
}

sub view_head2 {
   my ($self, $item) = @_;
   "<h2>", $item->title->present ($self), "</h2>",
   $item->content->present ($self)
}

sub view_textblock {
   my ($self, $item) = @_;
   "<p>$item</p>"
}

sub view_over {
   my ($self, $item) = @_;
   "<ul>",
   $item->content->present ($self),
   "</ul>"
}

sub view_item {
   my ($self, $item) = @_;
   "<li>", ($item->title eq '*' ? () : '<b>' . $item->title . '</b>'),
   $item->content->present ($self), "</li>\n"
}

sub view_seq_text { $_[1] }

sub view_seq_link {
   my ($title, $href) = split /\|/, $_[1], 2;
   "<a href='$href'>$title</a>"
}


=head1 AUTHOR

Copyright © 2007 Pippijn van Steenhoven

=head1 LICENSE

This library is free software, you can redistribute it and/or modify
it under the same terms as Perl itself.

=cut

1
